# Copyright (c) 2013 by Ladislav Lhotka, CZ.NIC <lhotka@nic.cz>
#                       Martin Bjorklund <mbj@tail-f.com>
#
# Translator of YANG to the hybrid DSDL schema (see RFC 6110).
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Translator from YANG to hybrid DSDL schema.

It is designed as a plugin for the pyang program and defines several
new command-line options:

--dsdl-no-documentation
    No output of DTD compatibility documentation annotations

--dsdl-no-dublin-core
    No output of Dublin Core annotations

--dsdl-record-defs
    Record all top-level defs, even if they are not used

Three classes are defined in this module:

* `DSDLPlugin`: pyang plugin interface class

* `HybridDSDLSchema`: provides instance that performs the mapping
  of input YANG modules to the hybrid DSDL schema.

* `Patch`: utility class representing a patch to the YANG tree
  where augment and refine statements are recorded.
"""

__docformat__ = "reStructuredText"

import sys
import optparse
import time


from pyang import plugin, error, xpath_lexer, util, statements, types

from .schemanode import SchemaNode

def pyang_plugin_init():
    plugin.register_plugin(DSDLPlugin())

class DSDLPlugin(plugin.PyangPlugin):
    def add_output_format(self, fmts):
        self.multiple_modules = True
        fmts['dsdl'] = self
    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--dsdl-no-documentation",
                                 dest="dsdl_no_documentation",
                                 action="store_true",
                                 default=False,
                                 help="No output of DTD compatibility"
                                 " documentation annotations"),
            optparse.make_option("--dsdl-no-dublin-core",
                                 dest="dsdl_no_dublin_core",
                                 action="store_true",
                                 default=False,
                                 help="No output of Dublin Core"
                                 " metadata annotations"),
            optparse.make_option("--dsdl-record-defs",
                                 dest="dsdl_record_defs",
                                 action="store_true",
                                 default=False,
                                 help="Record all top-level defs"
                                 " (even if not used)"),
            optparse.make_option("--dsdl-lax-yang-version",
                                 dest="dsdl_lax_yang_version",
                                 action="store_true",
                                 default=False,
                                 help="Try to translate modules with "
                                 "unsupported YANG versions (use at own risk)"),
            ]
        g = optparser.add_option_group("Hybrid DSDL schema "
                                       "output specific options")
        g.add_options(optlist)

    def emit(self, ctx, modules, fd):
        if 'submodule' in [ m.keyword for m in modules ]:
            raise error.EmitError("Cannot translate submodules")
        emit_dsdl(ctx, modules, fd)

def emit_dsdl(ctx, modules, fd):
    for (epos, etag, eargs) in ctx.errors:
        if error.is_error(error.err_level(etag)):
            raise error.EmitError("DSDL translation needs a valid module")
    schema = HybridDSDLSchema().from_modules(modules,
                                  ctx.opts.dsdl_no_dublin_core,
                                  ctx.opts.dsdl_no_documentation,
                                  ctx.opts.dsdl_record_defs,
                                  ctx.opts.dsdl_lax_yang_version,
                                  debug=0)
    fd.write(schema.serialize())

class Patch(object):

    """Instances of this class represent a patch to the YANG tree.

    A Patch is filled with substatements of 'refine' and/or 'augment'
    that are to be applied to a single node.

    Instance variables:

    * `self.path`: list specifying the relative path to the node where
      the patch is to be applied

    * `self.plist`: list of statements to apply
    """

    def __init__(self, path, refaug):
        """Initialize the instance with `refaug` statement.

        `refaug` must be either 'refine' or 'augment'.
        """
        self.path = path
        self.plist = [refaug]

    def pop(self):
        """Pop and return the first element of `self.path`."""
        return self.path.pop(0)

    def combine(self, patch):
        """Add `patch.plist` to `self.plist`."""
        exclusive = set(["config", "default", "mandatory", "presence",
                     "min-elements", "max-elements"])
        kws = set([s.keyword for s in self.plist]) & exclusive
        add = [n for n in patch.plist if n.keyword not in kws]
        self.plist.extend(add)

class HybridDSDLSchema(object):

    """Instance of this class maps YANG to the hybrid DSDL schema.

    Typically, only a single instance is created.

    Instance variables:

    * `self.all_defs`: dictionary of all named pattern
      definitions. The keys are mangled names of the definitions.

    * `self.data`: root of the data tree.

    * `self.debug`: debugging information level (0 = no debugging).

    * `self.gg_level`: level of immersion in global groupings.

    * `self.global_defs`: dictionary of global (aka chameleon) named
      pattern definitions. The keys are mangled names of the
      definitions.

    * `self.identities`: dictionary of identity names as keys and the
      corresponding name pattern definitions as values.

    * `self.identity_deps: each item has an identity (statement) as
      the key and a list of identities derived from the key identity
      as the value.

    * `self.local_defs`: dictionary of local named pattern
      definitions. The keys are mangled names of the definitions.

    * `self.local_grammar`: the inner <grammar> element containing the
      mapping of a single YANG module.

    * `self.module`: the module being processed.

    * `self.module_prefixes`: maps module names to (disambiguated)
      prefixes.

    * `self.namespaces`: maps used namespace URIs to (disambiguated)
      prefixes.

    * `self.notifications`: root of the subtree containing
      notifications.

    * `self.prefix_stack`: stack of active module prefixes. A new
      prefix is pushed on the stack for an augment from an external
      module.

    * `self.rpcs`: root of the subtree containing RPC signatures.

    * `self.stmt_handler`: dictionary of methods that are dispatched
      for handling individual YANG statements. Its keys are YANG
      statement keywords.

    * `self.top_grammar`: the outer (root) <grammar> element.

    * `self.type_handler`: dictionary of methods that are dispatched
      for handling individual YANG types. Its keys are the names of
      YANG built-in types.

    * `self.tree`: outer <start> pattern.

    """

    YANG_version = 1.1
    """Checked against the yang-version statement, if present."""

    dc_uri = "http://purl.org/dc/terms"
    """Dublin Core URI"""
    a_uri =  "http://relaxng.org/ns/compatibility/annotations/1.0"
    """DTD compatibility annotations URI"""

    datatype_map = {
        "int8": "byte",
        "int16": "short",
        "int32": "int",
        "int64": "long",
        "uint8": "unsignedByte",
        "uint16": "unsignedShort",
        "uint32": "unsignedInt",
        "uint64": "unsignedLong",
        "decimal64": "decimal",
        "binary": "base64Binary",
        "string": "string",
    }
    """Mapping of simple datatypes from YANG to W3C datatype library"""

    data_nodes = ("leaf", "container", "leaf-list", "list",
                  "anydata", "anyxml", "rpc", "notification")
    """Keywords of YANG data nodes."""

    schema_nodes = data_nodes + ("choice", "case")
    """Keywords of YANG schema nodes."""

    def __init__(self):
        """Initialize the dispatch dictionaries."""
        self.stmt_handler = {
            "action": self.noop,
            "anyxml": self.anyxml_stmt,
            "anydata": self.anyxml_stmt,
            "argument": self.noop,
            "augment": self.noop,
            "base": self.noop,
            "belongs-to": self.noop,
            "bit": self.noop,
            "case": self.case_stmt,
            "choice": self.choice_stmt,
            "config": self.nma_attribute,
            "contact": self.noop,
            "container": self.container_stmt,
            "default": self.noop,
            "deviation": self.noop,
            "deviate": self.noop,
            "description": self.description_stmt,
            "enum" : self.enum_stmt,
            "error-app-tag": self.noop,
            "error-message": self.noop,
            "extension": self.noop,
            "feature": self.noop,
            "fraction-digits": self.noop,
            "identity": self.noop,
            "if-feature": self.noop,
            "import" : self.noop,
            "include" : self.include_stmt,
            "input": self.noop,
            "grouping" : self.noop,
            "key": self.noop,
            "leaf": self.leaf_stmt,
            "leaf-list": self.leaf_list_stmt,
            "length": self.noop,
            "list": self.list_stmt,
            "mandatory": self.noop,
            "max-elements": self.noop,
            "min-elements": self.noop,
            "modifier": self.noop,
            "module": self.noop,
            "must": self.must_stmt,
            "namespace": self.noop,
            "notification": self.notification_stmt,
            "ordered-by": self.nma_attribute,
            "organization": self.noop,
            "output": self.noop,
            "path": self.noop,
            "pattern": self.noop,
            "position": self.noop,
            "prefix": self.noop,
            "presence": self.noop,
            "range": self.noop,
            "reference": self.reference_stmt,
            "refine": self.noop,
            "require-instance": self.noop,
            "revision": self.noop,
            "revision-date": self.noop,
            "rpc": self.rpc_stmt,
            "status": self.nma_attribute,
            "submodule": self.noop,
            "type": self.type_stmt,
            "typedef" : self.noop,
            "unique" : self.unique_stmt,
            "units" : self.nma_attribute,
            "uses" : self.uses_stmt,
            "value": self.noop,
            "when" : self.when_stmt,
            "yang-version": self.yang_version_stmt,
            "yin-element": self.noop,
        }
        self.ext_handler = {
            "ietf-yang-metadata": {
                "annotation": self.noop
            }
        }
        self.type_handler = {
            "boolean": self.boolean_type,
            "binary": self.binary_type,
            "bits": self.bits_type,
            "decimal64": self.numeric_type,
            "enumeration": self.choice_type,
            "empty": self.noop,
            "identityref": self.identityref_type,
            "instance-identifier": self.instance_identifier_type,
            "int8": self.numeric_type,
            "int16": self.numeric_type,
            "int32": self.numeric_type,
            "int64": self.numeric_type,
            "leafref": self.leafref_type,
            "string" : self.string_type,
            "uint8": self.numeric_type,
            "uint16": self.numeric_type,
            "uint32": self.numeric_type,
            "uint64": self.numeric_type,
            "union": self.choice_type,
        }

    def serialize(self):
        """Return the string representation of the receiver."""
        res = '<?xml version="1.0" encoding="UTF-8"?>'
        for ns in self.namespaces:
            self.top_grammar.attr["xmlns:" + self.namespaces[ns]] = ns
        res += self.top_grammar.start_tag()
        for ch in self.top_grammar.children:
            res += ch.serialize()
        res += self.tree.serialize()
        for d in self.global_defs:
            res += self.global_defs[d].serialize()
        for i in self.identities:
            res += self.identities[i].serialize()
        return res + self.top_grammar.end_tag()

    def from_modules(self, modules, no_dc=False, no_a=False,
                     record_defs=False, lax_yang_version=False, debug=0):
        """Return the instance representing mapped input modules."""
        self.namespaces = {
            "urn:ietf:params:xml:ns:netmod:dsdl-annotations:1" : "nma",
        }
        if not no_dc: self.namespaces[self.dc_uri] = "dc"
        if not no_a: self.namespaces[self.a_uri] = "a"
        self.global_defs = {}
        self.all_defs = {}
        self.identity_deps = {}
        self.identities = {}
        self.debug = debug
        self.module_prefixes = {}
        gpset = {}
        self.gg_level = 0
        metadata = []
        self.has_meta = False
        for module in modules[0].i_ctx.modules.values():
            yver = module.search_one("yang-version")
            if yver and float(yver.arg) > 1.0 and not lax_yang_version:
                raise error.EmitError(
                    "DSDL plugin supports only YANG version 1.")
            if module.keyword == "module":
                for idn in module.i_identities.values():
                    self.register_identity(idn)
        for module in modules:
            self.add_namespace(module)
            self.module = module
            annots = module.search(("ietf-yang-metadata", "annotation"))
            for ann in annots:
                aname = (self.module_prefixes[ann.main_module().arg] + ":" +
                         ann.arg)
                optel = SchemaNode("optional")
                atel = SchemaNode("attribute", optel).set_attr("name", aname)
                self.handle_substmts(ann, atel)
                metadata.append(optel)
        if metadata:
            self.has_meta = True
            metel = SchemaNode.define("__yang_metadata__")
            self.global_defs["__yang_metadata__"] = metel
            for mattr in metadata:
                metel.subnode(mattr)
        for module in modules:
            self.module = module
            self.prefix_stack = [self.module_prefixes[module.arg]]
            for aug in module.search("augment"):
                self.add_patch(gpset, aug)
            for sub in [ module.i_ctx.get_module(inc.arg)
                         for inc in module.search("include") ]:
                for aug in sub.search("augment"):
                    self.add_patch(gpset, aug)
        self.setup_top()
        for module in modules:
            self.module = module
            self.local_defs = {}
            if record_defs: self.preload_defs()
            self.prefix_stack = [self.module_prefixes[module.arg]]
            self.create_roots(module)
            self.lookup_expand(module, list(gpset.keys()))
            self.handle_substmts(module, self.data, gpset)
            for d in list(self.local_defs.values()):
                self.local_grammar.subnode(d)
            self.tree.subnode(self.local_grammar)
            self.all_defs.update(self.local_defs)
        self.all_defs.update(self.global_defs)
        self.dc_element(self.top_grammar, "date", time.strftime("%Y-%m-%d"))
        return self

    def setup_top(self):
        """Create top-level elements of the hybrid schema."""
        self.top_grammar = SchemaNode("grammar")
        self.top_grammar.attr = {
            "xmlns": "http://relaxng.org/ns/structure/1.0",
            "datatypeLibrary": "http://www.w3.org/2001/XMLSchema-datatypes"}
        self.tree = SchemaNode("start")

    def create_roots(self, yam):
        """Create the top-level structure for module `yam`."""
        self.local_grammar = SchemaNode("grammar")
        self.local_grammar.attr = {
            "ns": yam.search_one("namespace").arg,
            "nma:module": self.module.arg}
        src_text = "YANG module '%s'" % yam.arg
        revs = yam.search("revision")
        if len(revs) > 0:
            src_text += " revision %s" % self.current_revision(revs)
        self.dc_element(self.local_grammar, "source", src_text)
        start = SchemaNode("start", self.local_grammar)
        self.data = SchemaNode("nma:data", start, interleave=True)
        self.data.occur = 2
        self.rpcs = SchemaNode("nma:rpcs", start, interleave=False)
        self.notifications = SchemaNode("nma:notifications", start,
                                        interleave=False)

    def yang_to_xpath(self, xpe):
        """Transform YANG's `xpath` to a form suitable for Schematron.

        1. Prefixes are added to unprefixed local names. Inside global
           groupings, the prefix is represented as the variable
           '$pref' which is substituted via Schematron abstract
           patterns.
        2. '$root' is prepended to every absolute location path.
        """
        if self.gg_level:
            pref = "$pref:"
        else:
            pref = self.prefix_stack[-1] + ":"
        toks = xpath_lexer.scan(xpe)
        prev = None
        res = ""
        for tok in toks:
            if (tok.type == "SLASH" and
                prev not in ("DOT", "DOTDOT", "RPAREN", "RBRACKET", "name",
                             "wildcard", "prefix_test")):
                res += "$root"
            elif tok.type == "name" and ":" not in tok.value:
                res += pref
            res += tok.value
            if tok.type != "_whitespace": prev = tok.type
        return res

    def add_namespace(self, module):
        """Add item uri:prefix for `module` to `self.namespaces`.

        The prefix to be actually used for `uri` is returned.  If the
        namespace is already present, the old prefix is used.  Prefix
        clashes are resolved by disambiguating `prefix`.
        """
        uri = module.search_one("namespace").arg
        prefix = module.search_one("prefix").arg
        if uri in self.namespaces: return self.namespaces[uri]
        end = 1
        new = prefix
        while new in list(self.namespaces.values()):
            new = "%s%x" % (prefix,end)
            end += 1
        self.namespaces[uri] = new
        self.module_prefixes[module.arg] = new
        for inc in module.search("include"):
            self.module_prefixes[inc.arg] = new
        return new

    def register_identity(self, id_stmt):
        """Register `id_stmt` with its base identity, if any.
        """
        bst = id_stmt.search_one("base")
        if bst:
            bder = self.identity_deps.setdefault(bst.i_identity, [])
            bder.append(id_stmt)

    def add_derived_identity(self, id_stmt):
        """Add pattern def for `id_stmt` and all derived identities.

        The corresponding "ref" pattern is returned.
        """
        p = self.add_namespace(id_stmt.main_module())
        if id_stmt not in self.identities:   # add named pattern def
            self.identities[id_stmt] = SchemaNode.define("__%s_%s" %
                                                         (p, id_stmt.arg))
            parent = self.identities[id_stmt]
            if id_stmt in self.identity_deps:
                parent = SchemaNode.choice(parent, occur=2)
                for i in self.identity_deps[id_stmt]:
                    parent.subnode(self.add_derived_identity(i))
            idval = SchemaNode("value", parent, p+":"+id_stmt.arg)
            idval.attr["type"] = "QName"
        res = SchemaNode("ref")
        res.attr["name"] = self.identities[id_stmt].attr["name"]
        return res

    def preload_defs(self):
        """Preload all top-level definitions."""
        for d in (self.module.search("grouping") +
                  self.module.search("typedef")):
            uname, dic = self.unique_def_name(d)
            self.install_def(uname, d, dic)

    def add_prefix(self, name, stmt):
        """Return `name` prepended with correct prefix.

        If the name is already prefixed, the prefix may be translated
        to the value obtained from `self.module_prefixes`.  Unmodified
        `name` is returned if we are inside a global grouping.
        """
        if self.gg_level: return name
        pref, colon, local = name.partition(":")
        if colon:
            return (self.module_prefixes[stmt.i_module.i_prefixes[pref][0]]
                    + ":" + local)
        else:
            return self.prefix_stack[-1] + ":" + pref

    def qname(self, stmt):
        """Return (prefixed) node name of `stmt`.

        The result is prefixed with the local prefix unless we are
        inside a global grouping.
        """
        if self.gg_level: return stmt.arg
        return self.prefix_stack[-1] + ":" + stmt.arg

    def dc_element(self, parent, name, text):
        """Add DC element `name` containing `text` to `parent`."""
        if self.dc_uri in self.namespaces:
            dcel = SchemaNode(self.namespaces[self.dc_uri] + ":" + name,
                              text=text)
            parent.children.insert(0,dcel)

    def get_default(self, stmt, refd):
        """Return default value for `stmt` node.

        `refd` is a dictionary of applicable refinements that is
        constructed in the `process_patches` method.
        """
        if refd["default"]:
                return refd["default"]
        defst = stmt.search_one("default")
        if defst:
            return defst.arg
        return None

    def unique_def_name(self, stmt, inrpc=False):
        """Mangle the name of `stmt` (typedef or grouping).

        Return the mangled name and dictionary where the definition is
        to be installed. The `inrpc` flag indicates when we are inside
        an RPC, in which case the name gets the "__rpc" suffix.
        """
        module = stmt.main_module()
        name = ""
        while True:
            pref = stmt.arg if stmt.arg else stmt.keyword
            name = "__" + pref + name
            if stmt.keyword == "grouping": name = "_" + name
            if stmt.parent.parent is None: break
            stmt = stmt.parent
        defs = (self.global_defs
                if stmt.keyword in ("grouping", "typedef")
                else self.local_defs)
        if inrpc: name += "__rpc"
        return (module.arg + name, defs)

    def add_patch(self, pset, augref):
        """Add patch corresponding to `augref` to `pset`.

        `augref` must be either 'augment' or 'refine' statement.
        """
        try:
            path = [ self.add_prefix(c, augref)
                     for c in augref.arg.split("/") if c ]
        except KeyError:
            # augment of a module that's not among input modules
            return
        car = path[0]
        patch = Patch(path[1:], augref)
        if car in pset:
            sel = [ x for x in pset[car] if patch.path == x.path ]
            if sel:
                sel[0].combine(patch)
            else:
                pset[car].append(patch)
        else:
            pset[car] = [patch]

    def apply_augments(self, auglist, p_elem, pset):
        """Handle substatements of augments from `auglist`.

        The augments are applied in the context of `p_elem`.  `pset`
        is a patch set containing patches that may be applicable to
        descendants.
        """
        for a in auglist:
            par = a.parent
            if a.search_one("when") is None:
                wel = p_elem
            else:
                if p_elem.interleave:
                    kw = "interleave"
                else:
                    kw = "group"
                wel = SchemaNode(kw, p_elem, interleave=p_elem.interleave)
                wel.occur = p_elem.occur
            if par.keyword == "uses":
                self.handle_substmts(a, wel, pset)
                continue
            if par.keyword == "submodule":
                mnam = par.i_including_modulename
            else:
                mnam = par.arg
            if self.prefix_stack[-1] == self.module_prefixes[mnam]:
                self.handle_substmts(a, wel, pset)
            else:
                self.prefix_stack.append(self.module_prefixes[mnam])
                self.handle_substmts(a, wel, pset)
                self.prefix_stack.pop()

    def current_revision(self, r_stmts):
        """Pick the most recent revision date.

        `r_stmts` is a list of 'revision' statements.
        """
        cur = max([[int(p) for p in r.arg.split("-")] for r in r_stmts])
        return "%4d-%02d-%02d" % tuple(cur)

    def insert_doc(self, p_elem, docstring):
        """Add <a:documentation> with `docstring` to `p_elem`."""
        dtag = self.namespaces[self.a_uri] + ":documentation"
        elem = SchemaNode(dtag, text=docstring)
        p_elem.annots.append(elem)

    def install_def(self, name, dstmt, def_map, interleave=False):
        """Install definition `name` into the appropriate dictionary.

        `dstmt` is the definition statement ('typedef' or 'grouping')
        that is to be mapped to a RELAX NG named pattern '<define
        name="`name`">'. `def_map` must be either `self.local_defs` or
        `self.global_defs`. `interleave` determines the interleave
        status inside the definition.
        """
        delem = SchemaNode.define(name, interleave=interleave)
        delem.attr["name"] = name
        def_map[name] = delem
        if def_map is self.global_defs: self.gg_level += 1
        self.handle_substmts(dstmt, delem)
        if def_map is self.global_defs: self.gg_level -= 1

    def rng_annotation(self, stmt, p_elem):
        """Append YIN representation of extension statement `stmt`."""
        ext = stmt.i_extension
        prf, extkw = stmt.raw_keyword
        (modname,rev)=stmt.i_module.i_prefixes[prf]
        prefix = self.add_namespace(
            statements.modulename_to_module(self.module,modname,rev))
        eel = SchemaNode(prefix + ":" + extkw, p_elem)
        argst = ext.search_one("argument")
        if argst:
            if argst.search_one("yin-element", "true"):
                SchemaNode(prefix + ":" + argst.arg, eel, stmt.arg)
            else:
                eel.attr[argst.arg] = stmt.arg
        self.handle_substmts(stmt, eel)

    def propagate_occur(self, node, value):
        """Propagate occurence `value` to `node` and its ancestors.

        Occurence values are defined and explained in the SchemaNode
        class.
        """
        while node.occur < value:
            node.occur = value
            if node.name == "define":
                break
            node = node.parent

    def process_patches(self, pset, stmt, elem, altname=None):
        """Process patches for data node `name` from `pset`.

        `stmt` provides the context in YANG and `elem` is the parent
        element in the output schema. Refinements adding documentation
        and changing the config status are immediately applied. 

        The returned tuple consists of:
        - a dictionary of refinements, in which keys are the keywords
          of the refinement statements and values are the new values
          of refined parameters.
        - a list of 'augment' statements that are to be applied
          directly under `elem`.
        - a new patch set containing patches applicable to
          substatements of `stmt`.
        """
        if altname:
            name = altname
        else:
            name = stmt.arg
        new_pset = {}
        augments = []
        refine_dict = dict.fromkeys(("presence", "default", "mandatory",
                                     "min-elements", "max-elements"))
        for p in pset.pop(self.add_prefix(name, stmt), []):
            if p.path:
                head = p.pop()
                if head in new_pset:
                    new_pset[head].append(p)
                else:
                    new_pset[head] = [p]
            else:
                for refaug in p.plist:
                    if refaug.keyword == "augment":
                        augments.append(refaug)
                    else:
                        for s in refaug.substmts:
                            if s.keyword == "description":
                                self.description_stmt(s, elem, None)
                            elif s.keyword == "reference":
                                self.reference_stmt(s, elem, None)
                            elif s.keyword == "must":
                                self.must_stmt(s, elem, None)
                            elif s.keyword == "config":
                                self.nma_attribute(s, elem)
                            elif refine_dict.get(s.keyword, False) is None:
                                refine_dict[s.keyword] = s.arg
        return (refine_dict, augments, new_pset)

    def get_minmax(self, stmt, refine_dict):
        """Return pair of (min,max)-elements values for `stmt`.

        `stmt` must be a 'list' or 'leaf-list'. Applicable refinements
        from `refine_dict` are also taken into account.
        """
        minel = refine_dict["min-elements"]
        maxel = refine_dict["max-elements"]
        if minel is None:
            minst = stmt.search_one("min-elements")
            if minst:
                minel = minst.arg
            else:
                minel = "0"
        if maxel is None:
            maxst = stmt.search_one("max-elements")
            if maxst:
                maxel = maxst.arg
        if maxel == "unbounded": maxel = None
        return (minel, maxel)

    def lookup_expand(self, stmt, names):
        """Find schema nodes under `stmt`, also in used groupings.

        `names` is a list with qualified names of the schema nodes to
        look up. All 'uses'/'grouping' pairs between `stmt` and found
        schema nodes are marked for expansion.
        """
        if not names: return []
        todo = [stmt]
        while todo:
            pst = todo.pop()
            for sub in pst.substmts:
                if sub.keyword in self.schema_nodes:
                    qname = self.qname(sub)
                    if qname in names:
                        names.remove(qname)
                        par = sub.parent
                        while hasattr(par,"d_ref"): # par must be grouping
                            par.d_ref.d_expand = True
                            par = par.d_ref.parent
                        if not names: return [] # all found
                elif sub.keyword == "uses":
                    g = sub.i_grouping
                    g.d_ref = sub
                    todo.append(g)
        return names

    def type_with_ranges(self, tchain, p_elem, rangekw, gen_data):
        """Handle types with 'range' or 'length' restrictions.

        `tchain` is the chain of type definitions from which the
        ranges may need to be extracted. `rangekw` is the statement
        keyword determining the range type (either 'range' or
        'length'). `gen_data` is a function that generates the
        output schema node (a RELAX NG <data> pattern).
        """
        ranges = self.get_ranges(tchain, rangekw)
        if not ranges: return p_elem.subnode(gen_data())
        if len(ranges) > 1:
            p_elem = SchemaNode.choice(p_elem)
            p_elem.occur = 2
        for r in ranges:
            d_elem = gen_data()
            for p in self.range_params(r, rangekw):
                d_elem.subnode(p)
            p_elem.subnode(d_elem)

    def get_ranges(self, tchain, kw):
        """Return list of ranges defined in `tchain`.

        `kw` is the statement keyword determining the type of the
        range, i.e. 'range' or 'length'. `tchain` is the chain of type
        definitions from which the resulting range is obtained.

        The returned value is a list of tuples containing the segments
        of the resulting range.
        """
        (lo, hi) = ("min", "max")
        ran = None
        for t in tchain:
            rstmt = t.search_one(kw)
            if rstmt is None: continue
            parts = [ p.strip() for p in rstmt.arg.split("|") ]
            ran = [ [ i.strip() for i in p.split("..") ] for p in parts ]
            if ran[0][0] != 'min': lo = ran[0][0]
            if ran[-1][-1] != 'max': hi = ran[-1][-1]
        if ran is None: return None
        if len(ran) == 1:
            return [(lo, hi)]
        else:
            return [(lo, ran[0][-1])] + ran[1:-1] + [(ran[-1][0], hi)]

    def range_params(self, ran, kw):
        """Return list of <param>s corresponding to range `ran`.

        `kw` is the statement keyword determining the type of the
        range, i.e. 'range' or 'length'. `ran` is the internal
        representation of a range as constructed by the `get_ranges`
        method.
        """
        if kw == "length":
            if ran[0][0] != "m" and (len(ran) == 1 or ran[0] == ran[1]):
                elem = SchemaNode("param").set_attr("name","length")
                elem.text = ran[0]
                return [elem]
            min_ = SchemaNode("param").set_attr("name","minLength")
            max_ = SchemaNode("param").set_attr("name","maxLength")
        else:
            if len(ran) == 1: ran *= 2 # duplicating the value
            min_ = SchemaNode("param").set_attr("name","minInclusive")
            max_ = SchemaNode("param").set_attr("name","maxInclusive")
        res = []
        if ran[0][0] != "m":
            elem = min_
            elem.text = ran[0]
            res.append(elem)
        if ran[1][0] != "m":
            elem = max_
            elem.text = ran[1]
            res.append(elem)
        return res

    def handle_stmt(self, stmt, p_elem, pset={}):
        """
        Run handler method for statement `stmt`.

        `p_elem` is the parent node in the output schema. `pset` is
        the current "patch set" - a dictionary with keys being QNames
        of schema nodes at the current level of hierarchy for which
        (or descendants thereof) any pending patches exist. The values
        are instances of the Patch class.

        All handler methods are defined below and must have the same
        arguments as this method. They should create the output schema
        fragment corresponding to `stmt`, apply all patches from
        `pset` belonging to `stmt`, insert the fragment under `p_elem`
        and perform all side effects as necessary.
        """
        if self.debug > 0:
            sys.stderr.write("Handling '%s %s'\n" %
                             (util.keyword_to_str(stmt.raw_keyword), stmt.arg))
        try:
            method = self.stmt_handler[stmt.keyword]
        except KeyError:
            if isinstance(stmt.keyword, tuple):
                try:
                    method = self.ext_handler[stmt.keyword[0]][stmt.keyword[1]]
                except KeyError:
                    method = self.rng_annotation
                method(stmt, p_elem)
                return
            else:
                raise error.EmitError(
                    "Unknown keyword %s - this should not happen.\n"
                    % stmt.keyword)
        method(stmt, p_elem, pset)

    def handle_substmts(self, stmt, p_elem, pset={}):
        """Handle all substatements of `stmt`."""
        for sub in stmt.substmts:
            self.handle_stmt(sub, p_elem, pset)

    # Handlers for YANG statements

    def noop(self, stmt, p_elem, pset=''):
        """`stmt` is not handled in the regular way."""
        pass

    def anyxml_stmt(self, stmt, p_elem, pset):
        elem = SchemaNode.element(self.qname(stmt), p_elem)
        if self.has_meta:
            elem.annot(
                SchemaNode("ref").set_attr("name", "__yang_metadata__"))
        SchemaNode("parentRef", elem).set_attr("name", "__anyxml__")
        refd = self.process_patches(pset, stmt, elem)[0]
        if p_elem.name == "choice":
            elem.occur = 3
        elif refd["mandatory"] or stmt.search_one("mandatory", "true"):
            elem.occur = 2
            self.propagate_occur(p_elem, 2)
        self.handle_substmts(stmt, elem)

    def nma_attribute(self, stmt, p_elem, pset=None):
        """Map `stmt` to a NETMOD-specific attribute.

        The name of the attribute is the same as the 'keyword' of
        `stmt`.
        """
        att = "nma:" + stmt.keyword
        if att not in p_elem.attr:
            p_elem.attr[att] = stmt.arg

    def case_stmt(self, stmt, p_elem, pset):
        celem = SchemaNode.case(p_elem)
        if p_elem.default_case != stmt.arg:
            celem.occur = 3
        refd, augs, new_pset = self.process_patches(pset, stmt, celem)
        left = self.lookup_expand(stmt, list(new_pset.keys()))
        for a in augs:
            left = self.lookup_expand(a, left)
        self.handle_substmts(stmt, celem, new_pset)
        self.apply_augments(augs, celem, new_pset)

    def choice_stmt(self, stmt, p_elem, pset):
        chelem = SchemaNode.choice(p_elem)
        chelem.attr["nma:name"] = stmt.arg
        refd, augs, new_pset = self.process_patches(pset, stmt, chelem)
        left = self.lookup_expand(stmt, list(new_pset.keys()))
        for a in augs:
            left = self.lookup_expand(a, left)
        if refd["mandatory"] or stmt.search_one("mandatory", "true"):
            chelem.attr["nma:mandatory"] = "true"
            self.propagate_occur(chelem, 2)
        else:
            defv = self.get_default(stmt, refd)
            if defv is not None:
                chelem.default_case = defv
            else:
                chelem.occur = 3
        self.handle_substmts(stmt, chelem, new_pset)
        self.apply_augments(augs, chelem, new_pset)
        
    def container_stmt(self, stmt, p_elem, pset):
        celem = SchemaNode.element(self.qname(stmt), p_elem)
        if self.has_meta:
            celem.annot(
                SchemaNode("ref").set_attr("name", "__yang_metadata__"))
        refd, augs, new_pset = self.process_patches(pset, stmt, celem)
        left = self.lookup_expand(stmt, list(new_pset.keys()))
        for a in augs:
            left = self.lookup_expand(a, left)
        if (p_elem.name == "choice" and p_elem.default_case != stmt.arg
            or p_elem.name == "case" and
            p_elem.parent.default_case != stmt.parent.arg and
            len(stmt.parent.i_children) < 2 or
            refd["presence"] or stmt.search_one("presence")):
            celem.occur = 3
        self.handle_substmts(stmt, celem, new_pset)
        self.apply_augments(augs, celem, new_pset)

    def description_stmt(self, stmt, p_elem, pset):
        # ignore imported and top-level descriptions + desc. of enum
        if (self.a_uri in self.namespaces and
            stmt.i_module == self.module != stmt.parent and
            stmt.parent.keyword != "enum"):
            self.insert_doc(p_elem, stmt.arg)

    def enum_stmt(self, stmt, p_elem, pset):
        elem = SchemaNode("value", p_elem, stmt.arg)
        for sub in stmt.search("status"):
            self.handle_stmt(sub, elem)

    def include_stmt(self, stmt, p_elem, pset):
        if stmt.parent.keyword == "module":
            subm = self.module.i_ctx.get_module(stmt.arg)
            self.handle_substmts(subm, p_elem, pset)

    def leaf_stmt(self, stmt, p_elem, pset):
        qname = self.qname(stmt)
        elem = SchemaNode.element(qname)
        if self.has_meta:
            elem.annot(
                SchemaNode("ref").set_attr("name", "__yang_metadata__"))
        if p_elem.name == "_list_" and qname in p_elem.keys:
            p_elem.keymap[qname] = elem
            elem.occur = 2
        else:
            p_elem.subnode(elem)
        refd = self.process_patches(pset, stmt, elem)[0]
        if (p_elem.name == "choice" and p_elem.default_case != stmt.arg or
            p_elem.name == "case" and
            p_elem.parent.default_case != stmt.parent.arg and
            len(stmt.parent.i_children) < 2):
                elem.occur = 3
        elif refd["mandatory"] or stmt.search_one("mandatory", "true"):
            self.propagate_occur(elem, 2)
        if elem.occur == 0:
            defv = self.get_default(stmt, refd)
            if defv is not None:
                elem.default = defv
                self.propagate_occur(elem, 1)
        self.handle_substmts(stmt, elem)

    def leaf_list_stmt(self, stmt, p_elem, pset):
        lelem = SchemaNode.leaf_list(self.qname(stmt), p_elem)
        lelem.attr["nma:leaf-list"] = "true"
        if self.has_meta:
            lelem.annot(
                SchemaNode("ref").set_attr("name", "__yang_metadata__"))
        refd = self.process_patches(pset, stmt, lelem)[0]
        lelem.minEl, lelem.maxEl = self.get_minmax(stmt, refd)
        if int(lelem.minEl) > 0: self.propagate_occur(p_elem, 2)
        self.handle_substmts(stmt, lelem)

    def list_stmt(self, stmt, p_elem, pset):
        lelem = SchemaNode.list(self.qname(stmt), p_elem)
        if self.has_meta:
            lelem.annot(
                SchemaNode("ref").set_attr("name", "__yang_metadata__"))
        keyst = stmt.search_one("key")
        if keyst: lelem.keys = [self.add_prefix(k, stmt)
                                for k in keyst.arg.split()]
        refd, augs, new_pset = self.process_patches(pset, stmt, lelem)
        left = self.lookup_expand(stmt, list(new_pset.keys()) + lelem.keys)
        for a in augs:
            left = self.lookup_expand(a, left)
        lelem.minEl, lelem.maxEl = self.get_minmax(stmt, refd)
        if int(lelem.minEl) > 0: self.propagate_occur(p_elem, 2)
        self.handle_substmts(stmt, lelem, new_pset)
        self.apply_augments(augs, lelem, new_pset)

    def must_stmt(self, stmt, p_elem, pset):
        mel = SchemaNode("nma:must")
        p_elem.annot(mel)
        mel.attr["assert"] = self.yang_to_xpath(stmt.arg)
        em = stmt.search_one("error-message")
        if em:
            SchemaNode("nma:error-message", mel, em.arg)
        eat = stmt.search_one("error-app-tag")
        if eat:
            SchemaNode("nma:error-app-tag", mel, eat.arg)

    def notification_stmt(self, stmt, p_elem, pset):
        notel = SchemaNode("nma:notification", self.notifications)
        notel.occur = 2
        elem = SchemaNode.element(self.qname(stmt), notel,
                                  interleave=True, occur=2)
        augs, new_pset = self.process_patches(pset, stmt, elem)[1:]
        self.handle_substmts(stmt, elem, new_pset)
        self.apply_augments(augs, elem, new_pset)

    def reference_stmt(self, stmt, p_elem, pset):
        # ignore imported and top-level descriptions + desc. of enum
        if (self.a_uri in self.namespaces and
            stmt.i_module == self.module != stmt.parent and
            stmt.parent.keyword != "enum"):
            self.insert_doc(p_elem, "See: " + stmt.arg)

    def rpc_stmt(self, stmt, p_elem, pset):
        rpcel = SchemaNode("nma:rpc", self.rpcs)
        r_pset = self.process_patches(pset, stmt, rpcel)[2]
        inpel = SchemaNode("nma:input", rpcel)
        elem = SchemaNode.element(self.qname(stmt), inpel, occur=2)
        augs, pset = self.process_patches(r_pset,stmt,elem,"input")[1:]
        inst = stmt.search_one("input")
        if inst:
            self.handle_substmts(inst, elem, pset)
        else:
            SchemaNode("empty", elem)
        self.apply_augments(augs, elem, pset)
        augs, pset = self.process_patches(r_pset,stmt,None,"output")[1:]
        oust = stmt.search_one("output")
        if oust or augs:
            outel = SchemaNode("nma:output", rpcel)
            outel.occur = 2
            if oust: self.handle_substmts(oust, outel, pset)
            self.apply_augments(augs, outel, pset)
        self.handle_substmts(stmt, rpcel, r_pset)

    def type_stmt(self, stmt, p_elem, pset):
        """Handle ``type`` statement.

        Built-in types are handled by one of the specific type
        callback methods defined below.
        """
        typedef = stmt.i_typedef
        if typedef and not stmt.i_is_derived: # just ref
            uname, dic = self.unique_def_name(typedef)
            if uname not in dic:
                self.install_def(uname, typedef, dic)
            SchemaNode("ref", p_elem).set_attr("name", uname)
            defst = typedef.search_one("default")
            if defst:
                dic[uname].default = defst.arg
                occur = 1
            else:
                occur = dic[uname].occur
            if occur > 0: self.propagate_occur(p_elem, occur)
            return
        chain = [stmt]
        tdefault = None
        while typedef:
            type_ = typedef.search_one("type")
            chain.insert(0, type_)
            if tdefault is None:
                tdef = typedef.search_one("default")
                if tdef:
                    tdefault = tdef.arg
            typedef = type_.i_typedef
        if tdefault and p_elem.occur == 0:
            p_elem.default = tdefault
            self.propagate_occur(p_elem, 1)
        self.type_handler[chain[0].arg](chain, p_elem)

    def unique_stmt(self, stmt, p_elem, pset):
        def addpref(nid):
            xpath_nodes = []
            child = stmt.parent
            for node in nid.split("/"):
                prefixed_name = self.add_prefix(node, stmt)
                node_name = prefixed_name
                if ":" in prefixed_name:
                    node_name = prefixed_name.split(":")[1]
                if child is not None:
                    child = statements.search_child(child.substmts,
                                                    child.i_module.i_modulename,
                                                    node_name)
                if child is None or child.keyword not in ["choice", "case"]:
                    xpath_nodes.append(prefixed_name)
            return "/".join(xpath_nodes)
        uel = SchemaNode("nma:unique")
        p_elem.annot(uel)
        uel.attr["tag"] = " ".join(
            [addpref(nid) for nid in stmt.arg.split()])

    def uses_stmt(self, stmt, p_elem, pset):
        expand = False
        grp = stmt.i_grouping
        for sub in stmt.substmts:
            if sub.keyword in ("refine", "augment"):
                expand = True
                self.add_patch(pset, sub)
        if expand:
            self.lookup_expand(grp, list(pset.keys()))
        elif len(self.prefix_stack) <= 1 and not(hasattr(stmt,"d_expand")):
            uname, dic = self.unique_def_name(stmt.i_grouping,
                                              not(p_elem.interleave))
            if uname not in dic:
                self.install_def(uname, stmt.i_grouping, dic,
                                 p_elem.interleave)
            elem = SchemaNode("ref", p_elem).set_attr("name", uname)
            occur = dic[uname].occur
            if occur > 0: self.propagate_occur(p_elem, occur)
            self.handle_substmts(stmt, elem)
            return
        self.handle_substmts(grp, p_elem, pset)

    def when_stmt(self, stmt, p_elem, pset=None):
        p_elem.attr["nma:when"] = self.yang_to_xpath(stmt.arg)

    def yang_version_stmt(self, stmt, p_elem, pset):
        if float(stmt.arg) > self.YANG_version:
            raise error.EmitError("Unsupported YANG version: %s" % stmt.arg)

    # Handlers for YANG types

    def binary_type(self, tchain, p_elem):
        def gen_data():
            return SchemaNode("data").set_attr("type", "base64Binary")
        self.type_with_ranges(tchain, p_elem, "length", gen_data)

    def bits_type(self, tchain, p_elem):
        elem = SchemaNode("list", p_elem)
        zom = SchemaNode("zeroOrMore", elem)
        choi = SchemaNode.choice(zom, occur=2)
        for bit in tchain[0].search("bit"):
            SchemaNode("value", choi, bit.arg)

    def boolean_type(self, tchain, p_elem):
        elem = SchemaNode.choice(p_elem, occur=2)
        SchemaNode("value", elem, "true")
        SchemaNode("value", elem, "false")

    def choice_type(self, tchain, p_elem):
        """Handle ``enumeration`` and ``union`` types."""
        elem = SchemaNode.choice(p_elem, occur=2)
        self.handle_substmts(tchain[0], elem)

    def empty_type(self, tchain, p_elem):
        SchemaNode("empty", p_elem)

    def identityref_type(self, tchain, p_elem):
        bid = tchain[0].search_one("base").i_identity
        if bid not in self.identity_deps:
            sys.stderr.write("%s: warning: identityref has empty value space\n"
                             % tchain[0].pos)
            p_elem.subnode(SchemaNode("notAllowed"))
            p_elem.occur = 0
            return
        der = self.identity_deps[bid]
        if len(der) > 1:
            p_elem = SchemaNode.choice(p_elem, occur=2)
        for i in der:
            p_elem.subnode(self.add_derived_identity(i))

    def instance_identifier_type(self, tchain, p_elem):
        SchemaNode("parentRef", p_elem).attr["name"] = "__instance-identifier__"
        ii = SchemaNode("nma:instance-identifier")
        p_elem.annot(ii)
        rinst = tchain[0].search_one("require-instance")
        if rinst: ii.attr["require-instance"] = rinst.arg

    def leafref_type(self, tchain, p_elem):
        typ = tchain[0]
        occur = p_elem.occur
        pathstr = typ.parent.i_leafref.i_expanded_path
        p_elem.attr["nma:leafref"] = self.yang_to_xpath(pathstr)
        while type(typ.i_type_spec) == types.PathTypeSpec:
            typ = typ.i_type_spec.i_target_node.search_one("type")
        self.handle_stmt(typ, p_elem)
        if occur == 0: p_elem.occur = 0

    def mapped_type(self, tchain, p_elem):
        """Handle types that are simply mapped to RELAX NG."""
        SchemaNode("data", p_elem).set_attr("type",
                                            self.datatype_map[tchain[0].arg])

    def numeric_type(self, tchain, p_elem):
        """Handle numeric types."""
        typ = tchain[0].arg
        def gen_data():
            elem = SchemaNode("data").set_attr("type", self.datatype_map[typ])
            if typ == "decimal64":
                fd = tchain[0].search_one("fraction-digits").arg
                SchemaNode("param",elem,"19").set_attr("name","totalDigits")
                SchemaNode("param",elem,fd).set_attr("name","fractionDigits")
            return elem
        self.type_with_ranges(tchain, p_elem, "range", gen_data)

    def string_type(self, tchain, p_elem):
        pels = []
        for t in tchain:
            for pst in t.search("pattern"):
                pels.append(SchemaNode("param",
                                       text=pst.arg).set_attr("name","pattern"))
        def get_data():
            elem = SchemaNode("data").set_attr("type", "string")
            for p in pels: elem.subnode(p)
            return elem
        self.type_with_ranges(tchain, p_elem, "length", get_data)



from . import xpath_lexer
from . import xpath_parser
from .error import err_add
from .util import prefix_to_module, search_data_node, data_node_up

core_functions = {
    'last': ([], 'number'),
    'position': ([], 'number'),
    'count': (['node-set'], 'number'),
    'id': (['object'], 'node-set'),
    'local-name': (['node-set', '?'], 'string'),
    'namespace-uri': (['node-set', '?'], 'string'),
    'name': (['node-set', '?'], 'string'),
    'string': (['object'], 'string'),
    'concat': (['string', 'string', '*'], 'string'),
    'starts-with': (['string', 'string'], 'boolean'),
    'contains': (['string', 'string'], 'boolean'),
    'substring-before': (['string', 'string'], 'string'),
    'substring-after': (['string', 'string'], 'string'),
    'substring': (['string', 'number', 'number', '?'], 'string'),
    'string-length': (['string', '?'], 'number'),
    'normalize-space': (['string', '?'], 'string'),
    'translate': (['string', 'string', 'string'], 'string'),
    'boolean': (['object'], 'boolean'),
    'not': (['boolean'], 'boolean'),
    'true': ([], 'boolean'),
    'false': ([], 'boolean'),
    'lang': (['string'], 'boolean'),
    'number': (['object'], 'number'),
    'sum': (['node-set'], 'number'),
    'floor': (['number'], 'number'),
    'ceiling': (['number'], 'number'),
    'round': (['number'], 'number'),
    }

yang_xpath_functions = {
    'current': ([], 'node-set')
    }

yang_1_1_xpath_functions = {
    'bit-is-set': (['node-set', 'string'], 'boolean'),
    'enum-value': (['string'], 'number'),
    'deref': (['node-set'], 'node-set'),
    'derived-from': (['node-set', 'string'], 'boolean'),
    'derived-from-or-self': (['node-set', 'string'], 'boolean'),
    're-match': (['string', 'string'], 'boolean'),
    }

extra_xpath_functions = {
    'deref': (['node-set'], 'node-set'), # pyang extension for 1.0
    }

def add_extra_xpath_function(name, input_params, output_param):
    extra_xpath_functions[name] = (input_params, output_param)

def add_prefix(prefix, s):
    "Add `prefix` to all unprefixed names in `s`"
    # tokenize the XPath expression
    toks = xpath_lexer.scan(s)
    # add default prefix to unprefixed names
    toks2 = [_add_prefix(prefix, tok) for tok in toks]
    # build a string of the patched expression
    ls = [x.value for x in toks2]
    return ''.join(ls)

def _add_prefix(prefix, tok):
    if tok.type == 'name':
        m = xpath_lexer.re_ncname.match(tok.value)
        if m.group(2) == None:
            tok.value = prefix + ':' + tok.value
    return tok

## TODO: validate must/when after deviate

# node is the initial context node or None if it is not known
def v_xpath(ctx, stmt, node):
    try:
        if hasattr(stmt, 'i_xpath') and stmt.i_xpath is not None:
            q = stmt.i_xpath
        else:
            q = xpath_parser.parse(stmt.arg)
            stmt.i_xpath = q
        chk_xpath_expr(ctx, stmt.i_orig_module, stmt.pos, node, node, q)
    except xpath_lexer.XPathError as e:
        err_add(ctx.errors, stmt.pos, 'XPATH_SYNTAX_ERROR', e.msg)
        stmt.i_xpath = None
    except SyntaxError as e:
        err_add(ctx.errors, stmt.pos, 'XPATH_SYNTAX_ERROR', e.msg)
        stmt.i_xpath = None

# mod is the (sub)module where the stmt is defined, which we use to
# resolve prefixes.
def chk_xpath_expr(ctx, mod, pos, initial, node, q):
    if type(q) == type([]):
        chk_xpath_path(ctx, mod, pos, initial, node, q)
    elif type(q) == type(()):
        if q[0] == 'absolute':
            chk_xpath_path(ctx, mod, pos, initial, 'root', q[1])
        elif q[0] == 'relative':
            chk_xpath_path(ctx, mod, pos, initial, node, q[1])
        elif q[0] == 'union':
            for qa in q[1]:
                chk_xpath_path(ctx, mod, pos, initial, node, qa)
        elif q[0] == 'comp':
            chk_xpath_expr(ctx, mod, pos, initial, node, q[2])
            chk_xpath_expr(ctx, mod, pos, initial, node, q[3])
        elif q[0] == 'arith':
            chk_xpath_expr(ctx, mod, pos, initial, node, q[2])
            chk_xpath_expr(ctx, mod, pos, initial, node, q[3])
        elif q[0] == 'bool':
            chk_xpath_expr(ctx, mod, pos, initial, node, q[2])
            chk_xpath_expr(ctx, mod, pos, initial, node, q[3])
        elif q[0] == 'negative':
            chk_xpath_expr(ctx, mod, pos, initial, node, q[1])
        elif q[0] == 'function_call':
            chk_xpath_function(ctx, mod, pos, initial, node, q[1], q[2])
        elif q[0] == 'path_expr':
            chk_xpath_expr(ctx, mod, pos, initial, node, q[1])
        elif q[0] == 'path': # q[1] == 'filter'
            chk_xpath_expr(ctx, mod, pos, initial, node, q[2])
            chk_xpath_expr(ctx, mod, pos, initial, node, q[3])
        elif q[0] == 'var':
            # NOTE: check if the variable is known; currently we don't
            # have any variables in YANG xpath expressions
            err_add(ctx.errors, pos, 'XPATH_VARIABLE', q[1])
        elif q[0] == 'literal':
            # kind of hack to detect qnames, and mark the prefixes
            # as being used in order to avoid warnings.
            s = q[1]
            if s[0] == s[-1] and s[0] in ("'", '"'):
                s = s[1:-1]
                i = s.find(':')
                # make sure there is just one : present
                if i != -1 and s[i + 1:].find(':') == -1:
                    prefix = s[:i]
                    # we don't want to report an error; just mark the
                    # prefix as being used.
                    my_errors = []
                    prefix_to_module(mod, prefix, pos, my_errors)
                    for (pos0, code, arg) in my_errors:
                        if code == 'PREFIX_NOT_DEFINED':
                            err_add(ctx.errors, pos0,
                                    'WPREFIX_NOT_DEFINED', arg)

def chk_xpath_function(ctx, mod, pos, initial, node, func, args):
    signature = None
    if func in core_functions:
        signature = core_functions[func]
    elif func in yang_xpath_functions:
        signature = yang_xpath_functions[func]
    elif (mod.i_version != '1' and func in yang_1_1_xpath_functions):
        signature = yang_1_1_xpath_functions[func]
    elif ctx.strict and func in extra_xpath_functions:
        err_add(ctx.errors, pos, 'STRICT_XPATH_FUNCTION', func)
        return None
    elif not(ctx.strict) and func in extra_xpath_functions:
        signature = extra_xpath_functions[func]

    if signature is None:
        err_add(ctx.errors, pos, 'XPATH_FUNCTION', func)
        return None

    # check that the number of arguments are correct
    nexp = len(signature[0])
    nargs = len(args)
    if (nexp == nargs):
        pass
    elif (nexp == 0 and nargs != 0):
        err_add(ctx.errors, pos, 'XPATH_FUNC_ARGS',
                (func, nexp, nargs))
    elif (signature[0][-1] == '?' and nargs == (nexp - 1)):
        pass
    elif signature[0][-1] == '?':
        err_add(ctx.errors, pos, 'XPATH_FUNC_ARGS',
                (func, "%s-%s" % (nexp-1, nexp), nargs))
    elif (signature[0][-1] == '*' and nargs >= (nexp - 1)):
        pass
    elif signature[0][-1] == '*':
        err_add(ctx.errors, pos, 'XPATH_FUNC_ARGS',
                (func, "at least %s" % (nexp-1), nargs))
    elif nexp != nargs:
        err_add(ctx.errors, pos, 'XPATH_FUNC_ARGS',
                (func, nexp, nargs))

    # FIXME implement checks from check_function()

    # check the arguments - FIXME check type
    for arg in args:
        chk_xpath_expr(ctx, mod, pos, initial, node, arg)
    return signature[1]

def chk_xpath_path(ctx, mod, pos, initial, node, path):
    if len(path) == 0:
        return
    head = path[0]
    if head[0] == 'var':
        # check if the variable is known as a node-set
        # currently we don't have any variables, so this fails
        err_add(ctx.errors, pos, 'XPATH_VARIABLE', head[1])
    elif head[0] == 'function_call':
        func = head[1]
        args = head[2]
        rettype = chk_xpath_function(ctx, mod, pos, initial, node, func, args)
        if rettype is not None:
            # known function, check that it returns a node set
            if rettype != 'node-set':
                err_add(ctx.errors, pos, 'XPATH_NODE_SET_FUNC', func)
        if func == 'current':
            chk_xpath_path(ctx, mod, pos, initial, initial, path[1:])
    elif head[0] == 'step':
        axis = head[1]
        nodetest = head[2]
        preds = head[3]
        node1 = None
        if axis == 'self':
            pass
        elif axis == 'child' and nodetest[0] == 'name':
            prefix = nodetest[1]
            name = nodetest[2]
            if prefix is None:
                if initial is None:
                    pmodule = None
                elif initial.keyword == 'module':
                    pmodule = initial
                else:
                    pmodule = initial.i_module
            else:
                pmodule = prefix_to_module(mod, prefix, pos, ctx.errors)
            # if node and initial are None, it means we're checking an XPath
            # expression when it is defined in a grouping or augment, i.e.,
            # when the full tree is not expanded.  in this case we can't check
            # the paths
            if pmodule is not None and node is not None and initial is not None:
                if node == 'root':
                    children = pmodule.i_children
                elif hasattr(node, 'i_children'):
                    children = node.i_children
                else:
                    children = []
                child = search_data_node(children, pmodule.i_modulename, name)
                if child is None and node == 'root':
                    err_add(ctx.errors, pos, 'XPATH_NODE_NOT_FOUND2',
                            (pmodule.i_modulename, name, pmodule.arg))
                elif child is None and node.i_module is not None:
                    err_add(ctx.errors, pos, 'XPATH_NODE_NOT_FOUND1',
                            (pmodule.i_modulename, name,
                             node.i_module.i_modulename, node.arg))
                elif child is None:
                    err_add(ctx.errors, pos, 'XPATH_NODE_NOT_FOUND2',
                            (pmodule.i_modulename, name, node.arg))
                elif (hasattr(initial, 'i_config') and initial.i_config
                      and hasattr(child, 'i_config') and not child.i_config):
                    err_add(ctx.errors, pos, 'XPATH_REF_CONFIG_FALSE',
                            (pmodule.i_modulename, name))
                else:
                    node1 = child
        elif axis == 'parent' and nodetest == ('node_type', 'node'):
            if node is None:
                pass
            elif node == 'root':
                err_add(ctx.errors, pos, 'XPATH_PATH_TOO_MANY_UP', ())
            else:
                p = data_node_up(node)
                if p is None:
                    err_add(ctx.errors, pos, 'XPATH_PATH_TOO_MANY_UP', ())
                else:
                    node1 = p
        else:
            # we can't validate the steps on other axis, but we can validate
            # functions etc.
            pass
        for p in preds:
            chk_xpath_expr(ctx, mod, pos, initial, node1, p)
        chk_xpath_path(ctx, mod, pos, initial, node1, path[1:])



"""MEF usage guidelines plugin
See MEF Assigned Names and Numbers (MANN) at https://wiki.mef.net/display/MANN/MEF+Assigned+Names+and+Numbers
"""

import optparse
import sys

from pyang import plugin
from pyang import statements
from pyang import error
from pyang.error import err_add
from pyang.plugins import lint

def pyang_plugin_init():
    plugin.register_plugin(MEFPlugin())

class MEFPlugin(lint.LintPlugin):
    def __init__(self):
        lint.LintPlugin.__init__(self)
        self.namespace_prefixes = ['urn:mef:yang:', 'urn:mef:xid:']
        self.modulename_prefixes = ['mef']

    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--mef",
                                 dest="mef",
                                 action="store_true",
                                 help="Validate the module(s) according to " \
                                 "MEF rules."),
            ]
        optparser.add_options(optlist)

    def setup_ctx(self, ctx):
        if not ctx.opts.mef:
            return

        ctx.max_line_len = 70
        self._setup_ctx(ctx)



"""sid plugin

Plugin used to generate or update .sid files.
Please refer to [I-D.ietf-core-sid], [I-D.ietf-core-comi], [I-D.ietf-core-yang-cbor]
and [I-D.ietf-core-yang-library] for more information.

"""

import optparse
import sys
import json
import collections
import re
import os
import errno
from pyang import plugin
from pyang import util
from pyang import error
from collections import OrderedDict

def pyang_plugin_init():
    plugin.register_plugin(SidPlugin())

class SidPlugin(plugin.PyangPlugin):

    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--sid-help",
                                 dest="sid_help",
                                 action="store_true",
                                 help="Print help on automatic SID generation"),
            optparse.make_option("--sid-generate-file",
                                 action="store",
                                 type="string",
                                 dest="generate_sid_file",
                                 help="Generate a .sid file."),
            optparse.make_option("--sid-update-file",
                                 action="store",
                                 type="string",
                                 dest="update_sid_file",
                                 help="Generate a .sid file based on a previous .sid file."),
            optparse.make_option("--sid-check-file",
                                 action="store",
                                 type="string",
                                 dest="check_sid_file",
                                 help="Check the consistency between a .sid file and the .yang file(s)."),
            optparse.make_option("--sid-list",
                                 action="store_true",
                                 dest="list_sid",
                                 help="Print the list of SID."),
            optparse.make_option("--sid-registration-info",
                                 action="store_true",
                                 dest="sid_registration_info",
                                 help="Print the information required by the SID registry."),
            optparse.make_option("--sid-extra-range",
                                 action="store",
                                 type="string",
                                 dest="extra_sid_range",
                                 help="Add an extra SID range during a .sid file update."),
            ]

        g = optparser.add_option_group("SID file specific options")
        g.add_options(optlist)

    def setup_ctx(self, ctx):
        if ctx.opts.sid_help:
            print_help()
            sys.exit(0)

    def setup_fmt(self, ctx):
        ctx.implicit_errors = False

    def post_validate_ctx(self, ctx, modules):
        nbr_option_specified = 0
        if ctx.opts.generate_sid_file is not None:
            nbr_option_specified += 1
        if ctx.opts.update_sid_file is not None:
            nbr_option_specified += 1
        if ctx.opts.check_sid_file is not None:
            nbr_option_specified += 1
        if nbr_option_specified == 0:
            return
        if nbr_option_specified > 1:
            sys.stderr.write("Invalid option, only one process on .sid file can be requested.\n")
            return

        fatal_error = False
        for (epos, etag, eargs) in ctx.errors:
            if not error.is_warning(error.err_level(etag)):
                fatal_error = True

        if fatal_error or (ctx.errors != [] and ctx.opts.check_sid_file is not None):
            sys.stderr.write("Invalid YANG module\n")
            return

        sid_file = SidFile()

        if ctx.opts.sid_registration_info:
            sid_file.sid_registration_info = True

        if ctx.opts.generate_sid_file is not None:
            sid_file.range = ctx.opts.generate_sid_file
            sid_file.is_consistent = False
            sid_file.sid_file_created = True

        if ctx.opts.update_sid_file is not None:
            sid_file.input_file_name = ctx.opts.update_sid_file

        if ctx.opts.check_sid_file is not None:
            sid_file.input_file_name = ctx.opts.check_sid_file
            sid_file.check_consistency = True
            if not sid_file.sid_registration_info:
                print("Checking consistency of '%s'" % sid_file.input_file_name)

        if ctx.opts.extra_sid_range is not None:
            if ctx.opts.update_sid_file is not None:
                sid_file.extra_range = ctx.opts.extra_sid_range
            else:
                sys.stderr.write("An extra SID range can be specified only during a .sid file update.\n")
                return

        if ctx.opts.list_sid:
            sid_file.list_content = True

        try:
            sid_file.process_sid_file(modules[0])

        except SidParsingError as e:
            sys.stderr.write("ERROR, %s\n" % e.msg)
            sys.exit(1)
        except SidFileError as e:
            sys.stderr.write("ERROR in '%s', %s\n" % (sid_file.input_file_name, e.msg))
            sys.exit(1)
        except EnvironmentError as e:
            if e.errno == errno.ENOENT:
                sys.stderr.write("ERROR, file '%s' not found\n" % e.filename)
            else:
                sys.stderr.write("ERROR, in file '%s' " % e.filename)
            sys.exit(1)
        except ValueError as e:
            if hasattr(e, "lineno") and sid_file.input_file_name != "test16-bad-toaster@2009-11-20.sid":
                # Present only in python 3.5 and later, except json.decoder.JSONDecodeError
                sys.stderr.write("ERROR in '%s', line %d, column %d, %s\n" % (sid_file.input_file_name, e.lineno, e.colno, e.msg))
            else:
                sys.stderr.write("ERROR in '%s', invalid JSON content\n" % sid_file.input_file_name)
            sys.exit(1)

def print_help():
    print("""
YANG Schema Item iDentifiers (SID) are globally unique unsigned integers used
to identify YANG items. SIDs are used instead of names to save space in
constrained applications such as COREconf. This plugin is used to automatically
generate and updated .sid files used to persist and distribute SID assignments.


COMMANDS

pyang [--sid-list] --sid-generate-file {count | entry-point:size} yang-filename
pyang [--sid-list] --sid-update-file sid-filename yang-filename
      [--sid-extra-range {count | entry-point:size}]
pyang [--sid-list] --sid-check-file sid-filename yang-filename


OPTIONS

--generate-sid-file

  This option is used to generate a new .sid file from a YANG module.
  
  Two arguments are required to generate a .sid file; the SID range assigned to
  the YANG module and its definition file. The SID range specified is a
  sub-range within a range obtained from a registrar or a sub-range within the
  experimental range (i.e. 60000 to 99999). The SID range consists of the first
  SID of the range, followed by a colon, followed by the number of SID
  allocated to the YANG module. The filename consists of the module name,
  followed by an @ symbol, followed by the module revision, followed by the
  ".yang" extension.
           
  This example shows how to generate the file toaster@2009-11-20.sid.

  $ pyang --sid-generate-file 20000:100 toaster@2009-11-20.yang
  
--update-sid-file

  Each time new items are added to a YANG module by the introduction of a new
  revision of this module, its included sub-modules or imported modules, the
  associated .sid file need to be updated. This is done by using the
  --sid-update-file option.

  Two arguments are required to generate a .sid file for an updated YANG
  module; the previous .sid file generated for the YANG module and the
  definition file of the updated module. Both filenames follow the usual
  naming conversion consisting of the module name, followed by an @ symbol,
  followed by the module revision, followed by the extension.

  This example shows how to generate the file toaster@2009-12-28.sid based
  on the SIDs already present in toaster@2009-11-20.sid.

  $ pyang --sid-update-file toaster@2009-11-20.sid toaster@2009-12-28.yang
  
-- check-sid-file

  The --sid-check-file option can be used at any time to verify if a .sid file
  need to be updated.

  Two arguments are required to verify a .sid file; the filename of the .sid
  file to be checked and the corresponding definition file.

  For example:

  $ pyang --sid-check-file toaster@2009-12-28.sid toaster@2009-12-28.yang
  
--list_sid

  The --list_sid option can be used before any of the previous options to
  obtains the list of SIDs assigned or validated. For example:

  $ pyang --list-sid --sid-generate-file 20000:100 toaster@2009-11-20.yang
  
--extra-sid-range

  If needed, an extra SID range can be assigned to an existing YANG module
  during its update with the --sid-extra-range option.

  For example, this command generates the file toaster@2009-12-28.sid using
  the initial range(s) present in toaster@2009-11-20.sid and the extra range
  specified in the command line.

  $ pyang --sid-update-file toaster@2009-11-20.sid
          toaster@2009-12-28.yang --sid-extra-range 20100:100
  
count
  The number of SID required when generating or updating a .sid file can be
  computed by specifying "count" as SID range.

  For example:

  $ pyang --sid-generate-file count toaster@2009-11-20.yang
  or:

  $ pyang --sid-update-file toaster@2009-11-20.sid
          toaster@2009-12-28.yang --sid-extra-range count
"""
)

############################################################
class SidFileError(Exception):
    def __init__(self, msg=""):
        self.msg = msg

class SidParsingError(Exception):
    """raised by plugins to fail the emit() function"""
    def __init__(self, msg=""):
        self.msg = msg

############################################################
class SidFile:
    def __init__(self):
        self.sid_file_created = False
        self.is_consistent = True
        self.check_consistency = False
        self.list_content = False
        self.sid_registration_info = False
        self.input_file_name = None
        self.range = None
        self.extra_range = None
        self.count = False
        self.node_highest = 0
        self.content = OrderedDict()

    def process_sid_file(self, module):
        self.module_name = module.i_modulename
        self.module_revision = util.get_latest_revision(module)
        self.output_file_name = '%s@%s.sid' % (self.module_name, self.module_revision)

        if self.range is not None:
            if self.range == 'count':
                self.count = True
            else:
                self.set_sid_range(self.range)

        if self.input_file_name is not None:
            if self.input_file_name[-4:] != ".sid":
                raise SidParsingError("File '%s' is not a .sid file" % self.input_file_name)

            with open(self.input_file_name) as f:
                self.content = json.load(f, object_pairs_hook=collections.OrderedDict)
            self.upgrade_sid_file_format() # This function can be removed after a reasonable transition period.
            self.validate_key_and_value()
            self.validate_ovelaping_ranges()
            self.validate_sid()

        if self.extra_range is not None:
            if self.extra_range == 'count':
                self.count = True
            else:
                self.set_sid_range(self.extra_range)
                self.validate_ovelaping_ranges()

        self.set_module_information()
        self.collect_module_items(module)

        if self.range == 'count':
            number_of_unassigned_yang_items = self.number_of_unassigned_yang_items()
            print("\nThis YANG module requires %d SIDs." % number_of_unassigned_yang_items)
            return

        if self.extra_range == 'count':
            number_of_SIDs_allocated = self.number_of_SIDs_allocated()
            number_of_SIDs_used = self.number_of_SIDs_used()
            number_of_SIDs_available = number_of_SIDs_allocated - number_of_SIDs_used
            number_of_unassigned_yang_items = self.number_of_unassigned_yang_items()

            print("\nNumber of SIDs allocated to this module: %d" % number_of_SIDs_allocated)
            print("Number of SIDs required by this version: %d" % (number_of_SIDs_used + number_of_unassigned_yang_items))
            if (number_of_unassigned_yang_items > number_of_SIDs_available):
                print("\nAn extra range of at least %d SIDs is required to perform this update." % (number_of_unassigned_yang_items - number_of_SIDs_available))
            else:
                print("\nThe update of the .sid file can be performed using the currently available SIDs.")
            return

        self.sort_items()
        self.assign_sid()

        if self.list_content:
            self.list_all_items()
        else:
            self.list_deleted_items()

        if self.check_consistency:
            if self.is_consistent:
                if self.sid_registration_info:
                    self.print_registration_information(module)
                else:
                    print("\nCheck completed successfully")
            else:
                print("\nThe .sid file need to be updated.")
        else:
            if self.is_consistent:
                print("No .sid file generated, the current .sid file is already up to date.")
            else:
                self.generate_file()
                if self.sid_file_created:
                    print("\nFile %s created" % self.output_file_name)
                else:
                    print("\nFile %s updated" % self.output_file_name)

                print ("Number of SIDs available : %d" % self.number_of_SIDs_allocated())
                print ("Number of SIDs used : %d" % self.number_of_SIDs_used())


    ########################################################
    def set_sid_range(self, range):
        components = range.split(':')
        if len(components) != 2 or not re.match(r'\d+:\d+', range):
            raise SidParsingError("invalid range in argument, must be '<entry-point>:<size>'.")

        if not 'assignment-ranges' in self.content:
            self.content['assignment-ranges'] = []
        self.content['assignment-ranges'].append(OrderedDict([('entry-point', int(components[0])), ('size', int(components[1]))]))

    ########################################################
    # Set the 'module-name' and/or 'module-revision' in the .sid file if require
    def set_module_information(self):
        if 'module-name' not in self.content or self.module_name != self.content['module-name']:
            self.content['module-name'] = self.module_name
            if self.check_consistency == True:
                print("ERROR, Mismatch between the module name defined in the .sid file and the .yang file.")
                self.is_consistent = False

        if 'module-revision' not in self.content or self.module_revision != self.content['module-revision']:
            self.content['module-revision'] = self.module_revision
            if self.check_consistency == True:
                print("ERROR, Mismatch between the module revision defined in the .sid file and the .yang file.")
                self.is_consistent = False

    ########################################################
    # Verify the tag and data type of each .sid file JSON object
    def validate_key_and_value(self):
        assignment_ranges_absent = True
        module_name_absent = True
        module_revision_absent = True
        items_absent = True

        for key in self.content:
            if key == 'assignment-ranges':
                assignment_ranges_absent = False
                if type(self.content[key]) != list:
                    raise SidFileError("key 'assignment-ranges', invalid  value.")
                self.validate_ranges(self.content[key])
                continue

            if key == 'module-name':
                module_name_absent = False
                continue

            if key == 'module-revision':
                module_revision_absent = False
                continue

            if key == 'items':
                items_absent = False
                if type(self.content[key]) != list:
                    raise SidFileError("key 'items', invalid value.")
                self.validate_items(self.content[key])
                continue

            raise SidFileError("invalid field '%s'." % key)

        if module_name_absent:
            raise SidFileError("mandatory field 'module-name' not present")

        if module_revision_absent:
            raise SidFileError("mandatory field 'module-revision' not present")

        if assignment_ranges_absent:
            raise SidFileError("mandatory field 'assignment-ranges' not present")

        if items_absent:
            raise SidFileError("mandatory field 'items' not present")

    def validate_ranges(self, ranges):
        entry_point_absent = True
        size_absent = True

        for range in ranges:
            for key in range:
                if key == 'entry-point':
                    entry_point_absent = False
                    if type(range[key]) != int:
                        raise SidFileError("invalid 'entry-point' value '%s'." % range[key])
                    continue

                if key == 'size':
                    size_absent = False
                    if type(range[key]) != int:
                        raise SidFileError("invalid 'size' value '%s'." % range[key])
                    continue

                raise SidFileError("invalid key '%s'." % key)

        if entry_point_absent:
            raise SidFileError("mandatory field 'entry-point' not present")

        if size_absent:
            raise SidFileError("mandatory field 'size' not present")


    def validate_items(self, items):
        namespace_absent = True
        identifier_absent  = True
        sid_absent  = True
        for item in items:
            for key in item:
                if key == 'namespace':
                    namespace_absent = False
                    # Verify (type(item[key]) != str on both python 2 and 3
                    if not isinstance(item[key], ("".__class__, u"".__class__)) or not re.match(r'module$|identity$|feature$|data$', item[key]):
                        raise SidFileError("invalid 'namespace' value '%s'." % item[key])
                    continue

                elif key == 'identifier':
                    identifier_absent = False
                    # Verify (type(item[key]) != str on both python 2 and 3
                    if not isinstance(item[key], ("".__class__, u"".__class__)):
                        raise SidFileError("invalid 'identifier' value '%s'." % item[key])
                    continue

                elif key == 'sid':
                    sid_absent = False
                    if type(item[key]) != int:
                        raise SidFileError("invalid 'sid' value '%s'." % item[key])
                    continue

                raise SidFileError("invalid key '%s'." % key)

        if namespace_absent:
            raise SidFileError("mandatory field 'entry-point' not present")

        if identifier_absent:
            raise SidFileError("mandatory field 'entry-point' not present")

        if sid_absent:
            raise SidFileError("mandatory field 'entry-point' not present")

    ########################################################
    # Verify if each range defined in the .sid file is distinct
    def validate_ovelaping_ranges(self):
        if 'assignment-ranges' in self.content:
            l = len(self.content['assignment-ranges'])
            if l > 1:
                for i in range(l-1):
                    i_first_sid = self.content['assignment-ranges'][i]['entry-point']
                    i_last_sid = i_first_sid + self.content['assignment-ranges'][i]['size']
                    for j in range(i+1, l):
                        j_first_sid = self.content['assignment-ranges'][j]['entry-point']
                        j_last_sid = j_first_sid + self.content['assignment-ranges'][j]['size']
                        if (i_first_sid >= j_first_sid and i_first_sid < j_last_sid)  or (i_last_sid > j_first_sid and i_last_sid < j_last_sid):
                            raise SidFileError("overlapping ranges are not allowed.")

    ########################################################
    # Verify if each SID listed in items is in range and is not duplicate.
    def validate_sid(self):
        self.content['items'].sort(key=lambda item:item['sid'])
        last_sid = -1
        for item in self.content['items']:
            if self.out_of_ranges(item['sid']):
                raise SidFileError("'sid' %d not within 'assignment-ranges'" % item['sid'])
            if item['sid'] == last_sid:
                raise SidFileError("duplicated 'sid' value %d " % item['sid'])
                last_sid = item['sid']

    def out_of_ranges(self, sid):
        for range in self.content['assignment-ranges']:
            if sid >= range['entry-point'] and sid < range['entry-point'] + range['size']:
                return False
        return True

    ########################################################
    # Collection of items defined in .yang file(s)
    def collect_module_items(self, module):
        if 'items' not in self.content:
            self.content['items'] = []

        for item in self.content['items']:
            item['status'] = 'd' # Set to 'd' deleted, updated to 'o' if present in .yang file

        self.merge_item('module', self.module_name)

        for name in module.i_ctx.modules:
            if module.i_ctx.modules[name].keyword == 'submodule':
                self.merge_item('module', module.i_ctx.modules[name].arg)

        for feature in module.i_features:
                self.merge_item('feature', feature)

        for children in module.i_children:
            if children.keyword == 'leaf' or children.keyword == 'leaf-list' or children.keyword == 'anyxml' or children.keyword == 'anydata':
                self.merge_item('data', self.getPath(children))

            elif children.keyword == 'container' or children.keyword == 'list':
                self.merge_item('data', self.getPath(children))
                self.collect_inner_data_nodes(children.i_children)

            elif children.keyword == 'choice' or children.keyword == 'case':
                self.collect_inner_data_nodes(children.i_children)

            elif children.keyword == 'rpc':
                self.merge_item('data', "/%s:%s" % (self.module_name ,children.arg) )
                for statement in children.i_children:
                    if statement.keyword == 'input' or statement.keyword == 'output':
                        self.collect_inner_data_nodes(statement.i_children)

            elif children.keyword == 'notification':
                self.merge_item('data', "/%s:%s" % (self.module_name ,children.arg))
                self.collect_inner_data_nodes(children.i_children)

        for identity in module.i_identities:
                self.merge_item('identity', identity)

        for substmt in module.substmts:
            if substmt.keyword == 'augment':
                self.collect_in_substmts(substmt.substmts)
            elif hasattr(substmt, 'i_extension') and substmt.i_extension.arg ==  'yang-data':
                self.collect_in_substmts(substmt.substmts)

    def collect_inner_data_nodes(self, children):
        for statement in children:
            if statement.keyword == 'leaf' or statement.keyword == 'leaf-list' or statement.keyword == 'anyxml' or statement.keyword == 'anydata':
                self.merge_item('data', self.getPath(statement))

            elif statement.keyword == 'container' or statement.keyword == 'list':
                self.merge_item('data', self.getPath(statement))
                self.collect_inner_data_nodes(statement.i_children)

            elif statement.keyword == 'action':
                self.merge_item('data', self.getPath(statement))
                for children in statement.i_children:
                    if children.keyword == 'input' or children.keyword == 'output':
                        self.collect_inner_data_nodes(children.i_children)

            elif statement.keyword == 'notification':
                self.merge_item('data', self.getPath(statement))
                self.collect_inner_data_nodes(statement.i_children)

            elif statement.keyword == 'choice' or statement.keyword == 'case':
                self.collect_inner_data_nodes(statement.i_children)

    def collect_in_substmts(self, substmts):
        for statement in substmts:
            if statement.keyword == 'leaf' or statement.keyword == 'leaf-list' or statement.keyword == 'anyxml' or statement.keyword == 'anydata':
                self.merge_item('data', self.getPath(statement))

            elif statement.keyword == 'container' or statement.keyword == 'list':
                self.merge_item('data', self.getPath(statement))
                self.collect_in_substmts(statement.substmts)

            elif statement.keyword == 'choice' or statement.keyword == 'case':
                self.collect_in_substmts(statement.substmts)

            elif statement.keyword == 'uses':
                self.collect_inner_data_nodes(statement.i_grouping.i_children)

    def getPath(self, statement):
        path = ""

        while statement.i_module != None:
            if statement.keyword != "case" and statement.keyword != "choice" and statement.keyword != "grouping" and not (hasattr(statement, 'i_extension') and statement.i_extension.arg ==  'yang-data'):
                # Locate the data node parent
                parent = statement.parent
                while parent.i_module != None:
                    if parent.keyword == 'module' or parent.keyword == 'container' or parent.keyword == 'list' or parent.keyword == 'notification' or parent.keyword == 'rpc' or parent.keyword == 'action':
                        break
                    parent = parent.parent

                if parent.i_module == None or parent.i_module != statement.i_module:
                    path = "/" + statement.i_module.arg + ":" + statement.arg + path
                else:
                    path = "/" + statement.arg + path
            statement = statement.parent

        return path

    def merge_item(self, namespace, identifier):
        for item in self.content['items']:
            if (namespace == item['namespace'] and identifier == item['identifier']):
                item['status'] = 'o' # Item already assigned
                return
        self.content['items'].append(OrderedDict([('namespace', namespace), ('identifier', identifier), ('sid', -1), ('status', 'n')]))
        self.is_consistent = False

    ########################################################
    # Sort the items list by 'namespace' and 'identifier'
    def sort_items(self):
        self.content['items'].sort(key=lambda item:item['identifier'])
        self.content['items'].sort(key=lambda item:item['namespace'], reverse=True)

    ########################################################
    # Identifier assignment
    def assign_sid(self):
        sid = -1
        for i in range(len(self.content['items'])):
            if self.content['items'][i]['sid'] == -1:
                sid = self.get_next_sid(sid)
                while self.sid_used(sid):
                    sid = self.get_next_sid(sid)
                self.content['items'][i]['sid'] = sid

    def sid_used(self, sid):
        for i in range(len(self.content['items'])):
            if self.content['items'][i]['sid'] == sid:
                return True
        return False

    def get_next_sid(self, sid):
        global range_idx

        if sid == -1:
            range_idx = 0
            return self.content['assignment-ranges'][0]['entry-point']

        sid += 1
        if sid < self.content['assignment-ranges'][range_idx]['entry-point'] + self.content['assignment-ranges'][range_idx]['size']:
            return sid

        range_idx += 1
        if range_idx < len(self.content['assignment-ranges']):
            return self.content['assignment-ranges'][range_idx]['entry-point']

        unassigned_yang_items = self.number_of_unassigned_yang_items()
        raise SidParsingError("The current SID range(s) are exhausted, %d extra SID(s) are required, use the --sid-extra-range option to add a SID range to this YANG module." % unassigned_yang_items)

    ########################################################
    def list_all_items(self):
        definition_removed = False

        print("\nSID        Assigned to")
        print("---------  --------------------------------------------------")
        for item in self.content['items']:
            status = ""
            if item['status'] == 'n' and not self.sid_file_created:
                status = " (New)"
            if item['status'] == 'd':
                status = " (Remove)"
                definition_removed = True

            print("%-9s  %s %s%s" % (item['sid'], item['namespace'], item['identifier'], status))

        if definition_removed:
            print("\nWARNING, obsolete definitions should be defined as 'deprecated' or 'obsolete'.")

    ########################################################
    def list_deleted_items(self):
        definition_removed = False
        for item in self.content['items']:
            if item['status'] == 'd':
                print("WARNING, item '%s' have been deleted form the .yang files." % item['identifier'])
                definition_removed = True

        if definition_removed:
            print("Obsolete definitions MUST NOT be removed from YANG modules, see RFC 6020 section 10.")
            print("These definition(s) should be reintroduced with a 'deprecated' or 'obsolete' status.")

    ########################################################
    def generate_file(self):
        for item in self.content['items']:
            del item['status']

        if os.path.exists(self.output_file_name):
            os.remove(self.output_file_name)

        with open(self.output_file_name, 'w') as outfile:
            json.dump(self.content, outfile, indent=2)

    ########################################################
    def number_of_SIDs_allocated(self):
        size = 0
        for range in self.content['assignment-ranges']:
            size += range['size']
        return size


    def number_of_unassigned_yang_items(self):
        yang_item_unassigned = 0

        for i in range(len(self.content['items'])):
            if self.content['items'][i]['sid'] == -1:
                yang_item_unassigned += 1

        return yang_item_unassigned

    def number_of_SIDs_used(self):
        sid_used = 0

        for i in range(len(self.content['items'])):
            if self.content['items'][i]['sid'] != -1:
                sid_used += 1

        return sid_used

    def number_of_SIDs_used_in_range(self, entry_point, size):
        sid_used = 0
        next_entry_point = entry_point + size

        for i in range(len(self.content['items'])):
            sid = self.content['items'][i]['sid']
            if sid >= entry_point and sid < next_entry_point:
                 sid_used += 1

        return sid_used

    ########################################################
    def print_registration_information(self, module):
        info={
            'module_name' : self.module_name,
            'module_revision' : self.module_revision,
            'yang_file' : '%s@%s.yang' % (self.module_name, self.module_revision),
            'ranges' : [],
            'submodules' : []
        }

        for range in self.content['assignment-ranges']:
            info['ranges'].append({
                'entry_point' : range['entry-point'],
                'size' : range['size'],
                'used' : self.number_of_SIDs_used_in_range(range['entry-point'], range['size'])
            })

        for name in module.i_ctx.modules:
            if module.i_ctx.modules[name].keyword == 'submodule':
                info['submodules'].append('%s@%s.yang' % (module.i_ctx.modules[name].arg, module.i_ctx.modules[name].i_latest_revision))

        print(json.dumps(info, indent=2))

    ########################################################
    # Perform the conversion to the .sid file fromat introduced by [I-D.ietf-core-sid] version 3.
    # This method can be remove after the proper transition period.
    def upgrade_sid_file_format(self):
        if not 'items' in self.content or not self.content['items'] or not 'type' in self.content['items'][0]:
            return

        for item in self.content['items']:
            if 'type' in item:
                if item['type'] == 'Module':
                    item['namespace'] = 'module'
                    item['identifier'] = item['label']

                elif item['type'] == 'Submodule':
                    item['namespace'] = 'module'
                    item['identifier'] = item['label']

                elif item['type'] == 'feature':
                    item['namespace'] = 'feature'
                    item['identifier'] = item['label']

                elif item['type'] == 'identity':
                    item['namespace'] = 'identity'
                    p = item['label'].rfind('/') + 1
                    item['identifier'] = item['label'][p:]

                elif item['type'] == 'node':
                    item['namespace'] = 'data'
                    item['identifier'] = '/' + self.module_name + ':' + item['label'][1:]

                elif item['type'] == 'notification':
                    item['namespace'] = 'data'
                    item['identifier'] = '/' + self.module_name + ':' + item['label'][1:]

                elif item['type'] == 'rpc':
                    item['namespace'] = 'data'
                    item['identifier'] = '/' + self.module_name + ':' + item['label'][1:]

                elif item['type'] == 'action':
                    item['namespace'] = 'data'
                    item['identifier'] = '/' + self.module_name + ':' + item['label'][1:]

            item.pop('type')
            item.pop('label')



import xml.parsers.expat

NC_NS_URI ="urn:ietf:params:xml:ns:netconf:base:1.0"
CAPABILITIES = {
    "urn:ietf:params:xml:ns:netconf:base:1.0" : "base",
    "urn:ietf:params:netconf:base:1.1" : "base",
    "urn:ietf:params:netconf:capability:writable-running:1.0" : "writable-running",
    "urn:ietf:params:netconf:capability:candidate:1.0" : "candidate",
    "urn:ietf:params:netconf:capability:startup:1.0" : "startup",
    "urn:ietf:params:netconf:capability:url:1.0" : "url",
    "urn:ietf:params:netconf:capability:xpath:1.0" : "xpath",
    "urn:ietf:params:netconf:capability:notification:1.0" : "notification",
    "urn:ietf:params:netconf:capability:with-defaults:1.0" : "with-defaults",
    }

class Capability:

    def __init__(self, uri):
        self.parameters = {}
        if "?" in uri:
            id, pars = uri.split("?")
            self.parse_pars(pars)
        else:
            id = uri
        self.id = id

    def parse_pars(self,pars):
        for p in pars.split("&"):
            name, value=p.split("=")
            self.parameters[name] = value

class HelloParser:

    def __init__(self):
        self.capabilities = []
        self.depth = self.state = 0
        self.buffer = ""
        self.parser = xml.parsers.expat.ParserCreate(namespace_separator=' ')
        self.parser.CharacterDataHandler = self.handleCharData
        self.parser.StartElementHandler = self.handleStartElement
        self.parser.EndElementHandler = self.handleEndElement

    def handleCharData(self, data):
        if self.state == self.depth == 3:
            self.buffer += data

    def handleStartElement(self, data, attrs):
        ns_uri, tag = data.split()
        if ns_uri == NC_NS_URI:
            if self.state == self.depth == 0 and tag == "hello":
                self.state = 1
            elif self.state == self.depth == 1 and tag == "capabilities":
                self.state = 2
            elif self.state == self.depth == 2 and tag == "capability":
                self.state = 3
        self.depth += 1
            
    def handleEndElement(self, data):
        ns_uri, tag = data.split()
        if ns_uri == NC_NS_URI:
            if self.state == self.depth == 1 and tag == "hello":
                self.state = 0
            elif self.state == self.depth == 2 and tag == "capabilities":
                self.state = 1
            elif self.state == self.depth == 3 and tag == "capability":
                self.capabilities.append(Capability(self.buffer))
                self.buffer = ""
                self.state = 2
        self.depth -= 1

    def parse(self, fd):
        self.parser.ParseFile(fd)
        return self

    def yang_modules(self):
        """
        Return a list of advertised YANG module names with revisions.

        Avoid repeated modules.
        """
        res = {}
        for c in self.capabilities:
            m = c.parameters.get("module")
            if m is None or m in res: continue
            res[m] = c.parameters.get("revision")
        return res.items()

    def get_features(self, yam):
        """Return list of features declared for module `yam`."""
        mcap = [ c for c in self.capabilities
                 if c.parameters.get("module", None) == yam ][0]
        if not mcap.parameters.get("features"): return []
        return mcap.parameters["features"].split(",")

    def registered_capabilities(self):
        """Return dictionary of non-YANG capabilities.

        Only capabilities from the `CAPABILITIES` dictionary are taken
        into account.
        """
        return dict ([ (CAPABILITIES[c.id],c) for c in self.capabilities
                 if c.id in CAPABILITIES ])




"""XPath 1.0 parser

PLY-based parser to build an AST for an XPath 1.0 expression.

References are to rules in:
http://www.w3.org/TR/1999/REC-xpath-19991116
"""

from . import yacc
from . import xpath_lexer

def parse(s):
    return parser.parse(s, lexer = lexer, debug = False)

def pparse(s):
    try:
        return parse(s)
    except xpath_lexer.XPathError as e:
        print('ERROR: %s:%s: %s' % (e.line, e.pos, e.msg))
        return None
    except SyntaxError as e:
        print('ERROR: %s' % e.msg)
        return None

### Parser follows

## [1]
def p_location_path_1(p):
    'LocationPath : RelativeLocationPath'
    p[0] = ('relative', p[1])
def p_location_path_2(p):
    'LocationPath : AbsoluteLocationPath'
    p[0] = ('absolute', p[1])

## [2]
def p_abs_location_path_1(p):
    'AbsoluteLocationPath : SLASH RelativeLocationPath'
    p[0] = p[2]
def p_abs_location_path_2(p):
    'AbsoluteLocationPath : SLASH'
    p[0] = []
def p_abs_location_path_3(p):
    'AbsoluteLocationPath : AbbreviatedAbsoluteLocationPath'
    p[0] = p[1]

## [3]
def p_rel_location_path_1(p):
    'RelativeLocationPath : Step'
    p[0] = [p[1]]
def p_rel_location_path_2(p):
    'RelativeLocationPath : RelativeLocationPath SLASH Step'
    p[0] = list(p[1])
    p[0].append(p[3])
def p_rel_location_path_3(p):
    'RelativeLocationPath : AbbreviatedRelativeLocationPath'
    p[0] = p[1]

## [4]
def p_step_1(p):
    'Step : axis DOUBLECOLON NodeTest PredicateList'
    p[0] = ('step', p[1], p[3], p[4])
def p_step_2(p):
    'Step : axis DOUBLECOLON NodeTest'
    p[0] = ('step', p[1], p[3], [])
def p_step_3(p):
    'Step : AT name PredicateList'
    p[0] = ('step', 'attribute', p[2], p[3])
def p_step_4(p):
    'Step : AT name'
    p[0] = ('step', 'attribute', p[2], [])
def p_step_5(p):
    'Step : NodeTest PredicateList'
    p[0] = ('step', 'child', p[1], p[2])
def p_step_6(p):
    'Step : NodeTest'
    p[0] = ('step', 'child', p[1], [])
def p_step_7(p):
    'Step : AbbreviatedStep PredicateList'
    if p[1][1] == 'self':
        a = "."
        x = "self::node()"
    else:
        a = ".."
        x = "parent::node()"
    msg = "%s[<pred>] is illegal syntax.  use %s[<pred>] instead," % (a, x)
    raise xpath_lexer.XPathError(msg, 1, 1)
def p_step_8(p):
    'Step : AbbreviatedStep'
    p[0] = p[1]

def p_pred_list_1(p):
    'PredicateList : PredicateList Predicate'
    p[0] = list(p[1])
    p[0].append(p[2])
def p_pred_list_2(p):
    'PredicateList : Predicate'
    p[0] = [p[1]]

## [7]
def p_node_test_1(p):
    'NodeTest : NameTest'
    p[0] = p[1]
def p_node_test_2(p):
    'NodeTest : node_type LPAREN RPAREN'
    p[0] = ('node_type', p[1])
def p_node_test_3(p):
    'NodeTest : node_type LPAREN literal RPAREN'
    if p[1] != 'processing-instruction':
        raise SyntaxError
    p[0] = ('processing-instruction', p[3])

## [8]
def p_pred(p):
    'Predicate : LBRACKET PredicateExpr RBRACKET'
    p[0] = p[2]

## [9]
def p_pred_expr(p):
    'PredicateExpr : Expr'
    p[0] = p[1]

## [10]
def p_abbrev_abs_loc_path(p):
    'AbbreviatedAbsoluteLocationPath : DOUBLESLASH RelativeLocationPath'
    p[0] = [_expand_double_slash()]
    p[0].extend(p[2])

## [11]
def p_abbrev_rel_loc_path(p):
    'AbbreviatedRelativeLocationPath : RelativeLocationPath DOUBLESLASH Step'
    p[0] = list(p[1])
    p[0].append(_expand_double_slash())
    p[0].append(p[3])

## [12]
def p_abbrev_step_1(p):
    'AbbreviatedStep : DOT'
    p[0] = ('step', 'self', ('node_type', 'node'), [])
def p_abbrev_step_2(p):
    'AbbreviatedStep : DOTDOT'
    p[0] = ('step', 'parent', ('node_type', 'node'), [])

## [14]
def p_expr(p):
    'Expr : OrExpr'
    p[0] = p[1]

## [15]
def p_prim_expr_1(p):
    'PrimaryExpr : DOLLAR name'
    p[0] = ('variable', p[2])
def p_prim_expr_2(p):
    'PrimaryExpr : LPAREN Expr RPAREN'
    p[0] = p[2]
def p_prim_expr_3(p):
    'PrimaryExpr : literal'
    p[0] = ('literal', p[1])
def p_prim_expr_4(p):
    'PrimaryExpr : number'
    p[0] = ('number', p[1])
def p_prim_expr_5(p):
    'PrimaryExpr : FunctionCall'
    p[0] = p[1]

## [16]
def p_fun_call_1(p):
    'FunctionCall : function_name LPAREN RPAREN'
    p[0] = ('function_call', p[1], [])
def p_fun_call_2(p):
    'FunctionCall : function_name LPAREN ArgumentList RPAREN'
    p[0] = ('function_call', p[1], p[3])

def p_arg_list_1(p):
    'ArgumentList : ArgumentList COMMA Argument'
    p[0] = list(p[1])
    p[0].append(p[3])
def p_arg_list_2(p):
    'ArgumentList : Argument'
    p[0] = [p[1]]

## [17]
def p_arg(p):
    'Argument : Expr'
    p[0] = p[1]

## [18]
def p_union_expr_1(p):
    'UnionExpr : PathExpr'
    p[0] = p[1]
def p_union_expr_2(p):
    'UnionExpr : UnionExpr BAR PathExpr'
    p[0] = _mk_union(p[1], p[3])

## [19]
def p_path_expr_1(p):
    'PathExpr : LocationPath'
    p[0] = p[1]
def p_path_expr_2(p):
    'PathExpr : FilterExpr'
    p[0] = ('path_expr', p[1])
def p_path_expr_3(p):
    'PathExpr : FilterExpr SLASH RelativeLocationPath'
    p[0] = [p[1]]
    p[0].extend(p[3])
def p_path_expr_4(p):
    'PathExpr : FilterExpr DOUBLESLASH RelativeLocationPath'
    p[0] = [p[1]]
    p[0].append(_expand_double_slash())
    p[0].extend(p[3])

## [20]
def p_filter_expr_1(p):
    'FilterExpr : PrimaryExpr'
    p[0] = p[1]
def p_filter_expr_2(p):
    'FilterExpr : FilterExpr Predicate'
    p[0] = ('path', 'filter', (p[1], p[2]))

## [21]
def p_or_expr_1(p):
    'OrExpr : AndExpr'
    p[0] = p[1]
def p_or_expr_2(p):
    'OrExpr : OrExpr OR AndExpr'
    p[0] = ('bool', 'or', p[1], p[3])

## [22]
def p_and_expr_1(p):
    'AndExpr : EqualityExpr'
    p[0] = p[1]
def p_and_expr_2(p):
    'AndExpr : AndExpr AND EqualityExpr'
    p[0] = ('bool', 'and', p[1], p[3])

## [23]
def p_eq_expr_1(p):
    'EqualityExpr : RelationalExpr'
    p[0] = p[1]
def p_eq_expr_2(p):
    'EqualityExpr : EqualityExpr EQ RelationalExpr'
    p[0] = ('comp', '=', p[1], p[3])
def p_eq_expr_3(p):
    'EqualityExpr : EqualityExpr NEQ RelationalExpr'
    p[0] = ('comp', '!=', p[1], p[3])

##[24]
def p_rel_expr_1(p):
    'RelationalExpr : AdditiveExpr'
    p[0] = p[1]
def p_rel_expr_2(p):
    'RelationalExpr : RelationalExpr LT AdditiveExpr'
    p[0] = ('comp', '<', p[1], p[3])
def p_rel_expr_3(p):
    'RelationalExpr : RelationalExpr GT AdditiveExpr'
    p[0] = ('comp', '>', p[1], p[3])
def p_rel_expr_4(p):
    'RelationalExpr : RelationalExpr LTE AdditiveExpr'
    p[0] = ('comp', '<=', p[1], p[3])
def p_rel_expr_5(p):
    'RelationalExpr : RelationalExpr GTE AdditiveExpr'
    p[0] = ('comp', '>=', p[1], p[3])

## [25]
def p_add_expr_1(p):
    'AdditiveExpr : MultiplicativeExpr'
    p[0] = p[1]
def p_add_expr_2(p):
    'AdditiveExpr : AdditiveExpr PLUS MultiplicativeExpr'
    p[0] = ('arith', '+', p[1], p[3])
def p_add_expr_3(p):
    'AdditiveExpr : AdditiveExpr MINUS MultiplicativeExpr'
    p[0] = ('arith', '-', p[1], p[3])

## [26]
def p_mul_expr_1(p):
    'MultiplicativeExpr : UnaryExpr'
    p[0] = p[1]
def p_mul_expr_2(p):
    'MultiplicativeExpr : MultiplicativeExpr MultiplyOperator UnaryExpr'
    p[0] = ('arith', p[2], p[1], p[3])
def p_mul_expr_3(p):
    'MultiplicativeExpr : MultiplicativeExpr DIV UnaryExpr'
    p[0] = ('arith', 'div', p[1], p[3])
def p_mul_expr_4(p):
    'MultiplicativeExpr : MultiplicativeExpr MOD UnaryExpr'
    p[0] = ('arith', 'mod', p[1], p[3])

## [27]
def p_unary_expr_1(p):
    'UnaryExpr : UnionExpr'
    p[0] = p[1]
def p_unary_expr_2(p):
    'UnaryExpr : MINUS UnaryExpr %prec UMINUS'
    p[0] = ('negative', p[2])

## [34]
def p_mul_oper(p):
    'MultiplyOperator : STAR'
    p[0] = '*'

## [37]
def p_name_test_1(p):
    'NameTest : wildcard'
    p[0] = 'wildcard'
def p_name_test_2(p):
    'NameTest : prefix_test'
    p[0] = ('has_namespace', p[1])
def p_name_test_3(p):
    'NameTest : name'
    p[0] = _mk_name(p[1])

def p_error(tok):
    if tok:
        raise xpath_lexer.XPathError("syntax error before '%s'" % tok.value,
                                     tok.lineno, tok.lexpos)
    else:
        raise SyntaxError("unexpected end of string")

def _mk_union(a, b):
    if a[0] == 'union' and b[0] == 'union':
        v = list(a[1])
        v.extend(b[1])
        return ('union', v)
    elif a[0] == 'union':
        v = list(a[1])
        v.append(b[1])
        return ('union', v)
    elif b[0] == 'union':
        v = [b]
        v.extend(a[1])
        return ('union', v)
    else:
        return ('union', [a, b])

def _expand_double_slash():
    return ('step', 'descendant-or-self', ('node_type', 'node'), [])

def _mk_name(v):
    m = xpath_lexer.re_ncname.match(v)
    if m.group(2) is not None:
        return ('name', m.group(2), m.group(3))
    else:
        return ('name', None, v)

start = 'Expr'

precedence = (
    ('left', 'OR'),
    ('left', 'AND'),
    ('left', 'EQ', 'NEQ'),
    ('left', 'LT', 'LTE', 'GT', 'GTE'),
    ('right', 'UMINUS'), # Unary minus operator
)

tokens = xpath_lexer.token_defs()
lexer = xpath_lexer.XPathLexer()
parser = yacc.yacc(tabmodule="xpath_parsetab", debug=False)



"""IETF usage guidelines plugin
See RFC 8407
"""

import optparse
import sys
import re

from pyang import plugin
from pyang import statements
from pyang import error
from pyang.error import err_add
from pyang.plugins import lint

def pyang_plugin_init():
    plugin.register_plugin(IETFPlugin())

class IETFPlugin(lint.LintPlugin):
    def __init__(self):
        self.found_2119_keywords = False
        self.found_8174 = False
        self.found_tlp = False

        lint.LintPlugin.__init__(self)
        self.namespace_prefixes = ['urn:ietf:params:xml:ns:yang:']
        self.modulename_prefixes = ['ietf', 'iana']

    def add_opts(self, optparser):
        optlist = [
            optparse.make_option("--ietf",
                                 dest="ietf",
                                 action="store_true",
                                 help="Validate the module(s) according to " \
                                 "IETF rules."),
            optparse.make_option("--ietf-help",
                                 dest="ietf_help",
                                 action="store_true",
                                 help="Print help on the IETF checks and exit"),
            ]
        optparser.add_options(optlist)

    def setup_ctx(self, ctx):
        if ctx.opts.ietf_help:
            print_help()
            sys.exit(0)
        if not ctx.opts.ietf:
            return
        self._setup_ctx(ctx)

        statements.add_validation_fun(
            'grammar', ['description'],
            lambda ctx, s: self.v_chk_description(ctx, s))

        # register our error codes
        error.add_error_code(
            'IETF_MISSING_RFC8174', 4,
            'the module seems to use RFC 2119 keywords, but the required'
            + ' text from RFC 8174 is not found'
            + ' (see pyang --ietf-help for details).')

        error.add_error_code(
            'IETF_MISSING_TRUST_LEGAL_PROVISIONING', 4,
            'RFC 8407: 3.1: '
            + 'The IETF Trust Copyright statement seems to be'
            + ' missing (see pyang --ietf-help for details).')

    def v_chk_description(self, ctx, s):
        arg = re.sub(r'\s+', ' ', s.arg)
        if s.parent.keyword == 'module' or s.parent.keyword == 'submodule':
            m = re_rfc8174.search(arg)
            if m is not None:
                self.found_8174 = True
                arg = arg[:m.start()] + arg[m.end():]
            if re_tlp.search(arg) is None:
                err_add(ctx.errors, s.pos,
                        'IETF_MISSING_TRUST_LEGAL_PROVISIONING', ())
        if not self.found_2119_keywords:
            if re_2119_keywords.search(arg) is not None:
                self.found_2119_keywords = True
                if not self.found_8174:
                    err_add(ctx.errors, s.pos, 'IETF_MISSING_RFC8174', ())

def print_help():
    print("""
Validates the module or submodule according to the IETF rules found
in RFC 8407.

The module's or submodule's description statement must contain the
following text:

     Copyright (c) <year> IETF Trust and the persons identified as
     authors of the code.  All rights reserved.

     Redistribution and use in source and binary forms, with or
     without modification, is permitted pursuant to, and subject to
     the license terms contained in, the Simplified BSD License set
     forth in Section 4.c of the IETF Trust's Legal Provisions
     Relating to IETF Documents
     (https://trustee.ietf.org/license-info).

     This version of this YANG module is part of RFC XXXX
     (https://www.rfc-editor.org/info/rfcXXXX); see the RFC itself
     for full legal notices.

If any description statement in the module or submodule contains
RFC 2119 key words, the module's or submodule's description statement
must contain the following text:

     The key words 'MUST', 'MUST NOT', 'REQUIRED', 'SHALL', 'SHALL
     NOT', 'SHOULD', 'SHOULD NOT', 'RECOMMENDED', 'NOT RECOMMENDED',
     'MAY', and 'OPTIONAL' in this document are to be interpreted as
     described in BCP 14 (RFC 2119) (RFC 8174) when, and only when,
     they appear in all capitals, as shown here.
""")

rfc8174_str = \
r"""The key words 'MUST', 'MUST NOT', 'REQUIRED', 'SHALL', 'SHALL
NOT', 'SHOULD', 'SHOULD NOT', 'RECOMMENDED', 'NOT RECOMMENDED',
'MAY', and 'OPTIONAL' in this document are to be interpreted as
described in BCP 14 \(RFC 2119\) \(RFC 8174\) when, and only when,
they appear in all capitals, as shown here."""

re_rfc8174 = re.compile(re.sub(r'\s+', ' ', rfc8174_str))

tlp_str = \
r"""Copyright \(c\) [0-9]+ IETF Trust and the persons identified as
authors of the code\.  All rights reserved\.

Redistribution and use in source and binary forms, with or
without modification, is permitted pursuant to, and subject
to the license terms contained in, the Simplified BSD License
set forth in Section 4\.c of the IETF Trust's Legal Provisions
Relating to IETF Documents
\(https?://trustee.ietf.org/license-info\)\.

This version of this YANG module is part of
RFC .+(\s+\(https?://www.rfc-editor.org/info/rfc.+\))?; see
the RFC itself for full legal notices\."""

re_tlp = re.compile(re.sub(r'\s+', ' ', tlp_str))

re_2119_keywords = re.compile(r"MUST|SHOULD|SHALL|RECOMMENDED|MAY|OPTIONAL")




import copy

### struct to keep track of position for error messages

class Position(object):
    __slots__ = (
        'ref',
        'line',
        'top',
        'uses_pos',
    )

    def __init__(self, ref):
        self.ref = ref
        self.line = 0
        self.top = None
        self.uses_pos = None

    def __str__(self):
        s = self.ref + ':' + str(self.line)
        if self.uses_pos is None:
            return s
        else:
            return str(self.uses_pos) + ' (at ' + s + ')'

### Exceptions

class Abort(Exception):
    """used for non-recoverable errors to abort parsing"""
    pass

class Eof(Exception):
    """raised by tokenizer when end of file is detected"""
    pass

class TransformError(Exception):
    """raised by plugins to fail the transform() function"""

    def __init__(self, msg="", exit_code=1):
        self.msg = msg
        self.exit_code = exit_code

class EmitError(Exception):
    """raised by plugins to fail the emit() function"""
    def __init__(self, msg="", exit_code=1):
        self.msg = msg
        self.exit_code = exit_code

### error codes

## level:
##    1: critical error, can not be made into a warning
##    2: major error, can not be made into a warning
##    3: minor error, can be made into warning with -W
##    4: warning
error_codes = \
    {
    'READ_ERROR':
      (1,
       'read error: %s'),
    'EOF_ERROR':
      (1,
       'premature end of file'),
    'EXPECTED_QUOTED_STRING':
      (1,
       'expected quoted string after \'+\' operator'),
    'UNKNOWN_KEYWORD':
      (1,
       'unknown keyword "%s"'),
    'INCOMPLETE_STATEMENT':
      (1,
       'unterminated statement definition for keyword "%s", looking at %s'),
    'EXPECTED_KEYWORD':
      (1,
       'expected keyword "%s"'),
    'EXPECTED_KEYWORD_2':
      (1,
       'expected keyword "%s" as child to "%s"'),
    'EXPECTED_DATA_DEF':
      (1,
       'expected a data definition statement as child to "%s"'),
    'UNEXPECTED_KEYWORD':
      (1,
       'unexpected keyword "%s"'),
    'UNEXPECTED_KEYWORD_1':
      (1,
       'unexpected keyword "%s", expected "%s"'),
    'UNEXPECTED_KEYWORD_N':
      (1,
       'unexpected keyword "%s", expected one of %s'),
    'UNEXPECTED_KEYWORD_CANONICAL':
      (1,
       'keyword "%s" not in canonical order (see RFC 6020, Section 12)'),
    'UNEXPECTED_KEYWORD_CANONICAL_1':
      (1,
       'keyword "%s" not in canonical order, ' \
       'expected "%s" (see RFC 6020, Section 12)'),
    'UNEXPECTED_KEYWORD_USES':
      (1,
       'unexpected keyword "%s" under "%s", defined at %s'),
    'EXPECTED_ARGUMENT':
      (1,
       'expected an argument for keyword "%s"'),
    'UNEXPECTED_ARGUMENT':
      (1,
       'did not expect an argument, got "%s"'),
    'XML_IDENTIFIER':
      (3,
       'illegal identifier "%s", must not start with [xX][mM][lL] in' \
       ' YANG version 1 (see RFC 6020, Section 12)'),
    'TRAILING_GARBAGE':
      (2,
       'trailing garbage after module'),
    'BAD_VALUE':
      (1,
       'bad value "%s" (should be %s)'),
    'CIRCULAR_DEPENDENCY':
      (1,
       'circular dependency for %s "%s"'),
    'MODULE_NOT_FOUND':
      (1,
       'module "%s" not found in search path'),
    'MODULE_NOT_FOUND_REV':
      (1,
       'module "%s" revision "%s" not found in search path'),
    'MODULE_NOT_IMPORTED':
      (1,
       'no module with the namespace "%s" is imported'),
    'BAD_IMPORT':
      (1,
       'cannot import %s "%s", must be a module'),
    'BAD_IMPORT_YANG_VERSION':
      (1,
       'a version %s module cannot import a version %s module by revision'),
    'BAD_INCLUDE':
      (1,
       'cannot include %s "%s", must be a submodule'),
    'BAD_INCLUDE_YANG_VERSION':
      (1,
       'cannot include a version %s submodule in a version %s module'),
    'BAD_MODULE_NAME':
      (2,
       'unexpected modulename "%s" in %s, should be "%s"'),
    'WBAD_MODULE_NAME':
      (4,
       'unexpected modulename "%s" in %s, should be "%s"'),
    'FILENAME_BAD_MODULE_NAME':
      (4,
       'filename "%s" suggests invalid module name "%s", should match "%s"'),
    'BAD_REVISION':
      (3,
       'unexpected latest revision "%s" in %s, should be "%s"'),
    'WBAD_REVISION':
      (4,
       'unexpected latest revision "%s" in %s, should be "%s"'),
    'FILENAME_BAD_REVISION':
      (4,
       'filename "%s" suggests invalid revision "%s", should match "%s"'),
    'BAD_SUB_BELONGS_TO':
      (1,
       'module "%s" includes "%s", but "%s" does not specify a ' \
       'correct belongs-to'),
    'MISSING_INCLUDE':
      (1,
       'submodule %s is included by %s, but not by the module %s'),
    'PREFIX_ALREADY_USED':
      (1,
       'prefix "%s" already used for module %s'),
    'PREFIX_NOT_DEFINED':
      (1,
       'prefix "%s" is not defined (reported only once)'),
    'WPREFIX_NOT_DEFINED':
      (4,
       'prefix "%s" is not defined'),
    'NODE_NOT_FOUND':
      (1,
       'node %s::%s is not found'),
    'NODE_NOT_FOUND1':
      (1,
       'node %s::%s is not found in %s::%s'),
    'NODE_NOT_FOUND2':
      (1,
       'node %s::%s is not found in module %s'),
    'BAD_NODE_IN_AUGMENT':
      (1,
       'node %s::%s of type %s cannot be augmented'),
    'BAD_NODE_IN_REFINE':
      (1,
       'node %s::%s cannot be refined'),
    'BAD_REFINEMENT':
      (1,
       '"%s" node "%s::%s" cannot be refined with "%s"'),
    'BAD_DEVIATE_KEY':
      (2,
       'key node "%s::%s" cannot be deviated with "not-supported"'),
    'BAD_DEVIATE_ADD':
      (2,
       'the "%s" property already exists in node "%s::%s"'),
    'BAD_DEVIATE_DEL':
      (2,
       'the "%s" property does not exist in node "%s::%s"'),
    'BAD_DEVIATE_TYPE':
      (2,
       'the "%s" property cannot be added'),
    'BAD_DEVIATE_WITH_NOT_SUPPORTED':
      (2,
       'cannot have other deviate statement together with "not-supported"'),
    'EXTENSION_NOT_DEFINED':
      (1,
       'extension "%s" is not defined in module %s'),
    'TYPE_NOT_FOUND':
      (1,
       'type "%s" not found in module "%s"'),
    'FEATURE_NOT_FOUND':
      (1,
       'feature "%s" not found in module "%s"'),
    'IDENTITY_NOT_FOUND':
      (1,
       'identity "%s" not found in module "%s"'),
    'GROUPING_NOT_FOUND':
      (1,
       'grouping "%s" not found in module "%s"'),
    'DEFAULT_CASE_NOT_FOUND':
      (1,
       'the default case "%s" is not found"'),
    'MANDATORY_NODE_IN_DEFAULT_CASE':
      (1,
       'mandatory node in default case'),
    'MULTIPLE_REFINE':
      (1,
       'the node "%s" is already refined at %s'),
    'RANGE_BOUNDS':
      (2,
       'range error: "%s" is not larger than "%s"'),
    'LENGTH_BOUNDS':
      (2,
       'length error: "%s" is not larger than "%s"'),
    'LENGTH_VALUE':
      (2,
       'length error: "%s" is too large'),
    'TYPE_VALUE':
      (2,
       'the value "%s" does not match its base type %s- %s'),
    'DUPLICATE_ENUM_NAME':
      (1,
       'the enum name "%s" has already been used for the ' \
       'enumeration at %s'),
    'DUPLICATE_ENUM_VALUE':
      (1,
       'the integer value "%d" has already been used for the ' \
       'enumeration at %s'),
    'ENUM_VALUE':
      (1,
       'the enumeration value "%s" is not an 32 bit integer'),
    'BAD_ENUM_VALUE':
      (1,
       'the given value "%s" does not match the base enum value "%d"'),
    'DUPLICATE_BIT_POSITION':
      (1,
       'the position "%d" has already been used for the bit at %s'),
    'BIT_POSITION':
      (1,
       'the position value "%s" is not valid'),
    'BAD_BIT_POSITION':
      (1,
       'the given position "%s" does not match the base bit position "%d"'),
    'NEED_KEY':
      (1,
       'the list needs at least one key'),
    'NEED_KEY_USES':
      (1,
       'the list at "%s" needs at least one key because it is used as config'),
    'KEY_BAD_CONFIG':
      (1,
       'the key "%s" does not have same "config" as its list'),
    'BAD_KEY':
      (1,
       'the key "%s" does not reference an existing leaf'),
    'BAD_UNIQUE':
      (1,
       'the unique argument "%s" does not reference an existing leaf'),
    'BAD_UNIQUE_PART':
      (1,
       'the identifier "%s" in the unique argument does not reference '
       'an existing container'),
    'BAD_UNIQUE_PART_LIST':
      (1,
       'the identifier "%s" in the unique argument references a list; '
       'this is not legal'),
    'BAD_UNIQUE_CONFIG':
      (1,
       'the identifer "%s" has not the same config property as the'
       ' other nodes in the unique expression'),
    'ILLEGAL_ESCAPE':
      (1,
       'the escape sequence "\\%s" is illegal in double quoted strings'),
    'ILLEGAL_ESCAPE_WARN':
      (4,
       'the escape sequence "\\%s" is unsafe in double quoted strings' \
       ' - pass the flag --lax-quote-checks to avoid this warning'),
    'UNIQUE_IS_KEY':
      (4,
       'all keys in the list are redundantly present in the unique statement'),
    'DUPLICATE_KEY':
      (2,
       'the key "%s" must not be listed more than once'),
    'DUPLICATE_UNIQUE':
      (3,
       'the leaf "%s" occurs more than once in the unique expression'),
    'PATTERN_ERROR':
      (2,
       'syntax error in pattern: %s'),
    'PATTERN_FAILURE':
      (4,
       'could not verify pattern: %s'),
    'LEAFREF_TOO_MANY_UP':
      (1,
       'the path for %s at %s has too many ".."'),
    'LEAFREF_IDENTIFIER_NOT_FOUND':
      (1,
       '"%s:%s" in the path for %s at %s is not found'),
    'LEAFREF_IDENTIFIER_BAD_NODE':
      (1,
       '"%s:%s" in the path for %s at %s references a %s node'),
    'LEAFREF_BAD_PREDICATE':
      (1,
       '"%s:%s" in the path for %s at %s has a predicate, '
       'but is not a list'),
    'LEAFREF_BAD_PREDICATE_PTR':
      (1,
       '"%s:%s" in the path\'s predicate for %s at %s is compared '
       'with a node that is not a leaf'),
    'LEAFREF_NOT_LEAF':
      (1,
       'the path for %s at %s does not refer to a leaf'),
    'LEAFREF_NO_KEY':
      (1,
       '"%s:%s" in the path for %s at %s is not the name of a key leaf'),
    'LEAFREF_MULTIPLE_KEYS':
      (1,
       '"%s:%s" in the path for %s at %s is referenced more than once'),
    'LEAFREF_BAD_CONFIG':
      (1,
       'the path for %s is config but refers to a '
       'non-config leaf "%s" defined at %s'),
    'LEAFREF_DEREF_NOT_LEAFREF':
      (1,
       'the deref argument refers to node "%s" at %s which is'
       ' not a leafref leaf'),
    'LEAFREF_DEREF_NOT_KEY':
      (1,
       'the deref argument refers to node "%s" at %s which'
       ' does not refer to a key (%s at %s)'),
    'LEAFREF_TO_NOT_IMPLEMENTED':
      (1,
       'the leafref refer to a node that is not implemented'),
    'DUPLICATE_CHILD_NAME':
      (1,
       'there is already a child node to "%s" at %s with the name "%s" '
       'defined at %s'),
    'BAD_TYPE_NAME':
      (1,
       'illegal type name "%s"'),
    'TYPE_ALREADY_DEFINED':
      (1,
       'type name "%s" is already defined at %s'),
    'GROUPING_ALREADY_DEFINED':
      (1,
       'grouping name "%s" is already defined at %s'),
    'FEATURE_ALREADY_DEFINED':
      (1,
       'feature name "%s" is already defined at %s'),
    'IDENTITY_ALREADY_DEFINED':
      (1,
       'identity name "%s" is already defined at %s'),
    'EXTENSION_ALREADY_DEFINED':
      (1,
       'extension name "%s" is already defined at %s'),
    'BAD_RESTRICTION':
      (1,
       'restriction "%s" not allowed for this base type'),
    'BAD_DEFAULT_VALUE':
      (1,
       'the type "%s" cannot have a default value'),
    'MISSING_TYPE_SPEC':
      (1,
       'a type "%s" must have at least one "%s" statement'),
    'MISSING_TYPE_SPEC_1':
      (1,
       'a type "%s" must have a "%s" statement'),
    'BAD_TYPE_IN_UNION':
      (1,
       'the type "%s" (defined at %s) cannot be part of a union'),
    'BAD_TYPE_IN_KEY':
      (1,
       'the type "%s" cannot be part of a key, used by leaf "%s"'),
    'KEY_BAD_SUBSTMT':
      (1,
       'the statement "%s" cannot be given for a key'),
    'DEFAULT_AND_MANDATORY':
      (1,
       'a \'default\' value cannot be given when \'mandatory\' is "true"'),
    'DEFAULT_AND_MIN_ELEMENTS':
      (1,
       'a \'default\' value cannot be given when \'min-elements\' is'
       ' greater than 0'),
    'DUPLICATE_DEFAULT':
        (1,
         'the default value "%s" is given twice in the leaf list'),
    'BAD_STATUS_REFERENCE':
      (2,
       'the "%s" definition is %s, but the "%s" it references is %s'),
    'REVISION_ORDER':
      (4,
       'the revision statements are not given in reverse chronological order'),
    'EXTENSION_ARGUMENT_PRESENT':
      (1,
       'unexpected argument for extension "%s"'),
    'EXTENSION_NO_ARGUMENT_PRESENT':
      (1,
       'expected argument for extension "%s"'),
    'SYNTAX_ERROR':
      (1,
       'syntax error: %s'),
    'DUPLICATE_NAMESPACE':
      (1,
       'duplicate namespace uri "%s" found in module "%s"'),
    'MISSING_ARGUMENT_ATTRIBUTE':
      (1,
       'missing argument attribute "%s" for "%s"'),
    'MISSING_ARGUMENT_ELEMENT':
      (1,
       'missing argument element "%s" for "%s"'),
    'UNEXPECTED_ATTRIBUTE':
      (1,
       'unexpected attribute %s'),
    'INVALID_CONFIG':
      (2,
       'config true cannot be set when the parent is config false'),
    'XPATH_SYNTAX_ERROR':
      (2,
       'XPath syntax error: %s'),
    'XPATH_VARIABLE':
      (2,
       'XPath variable "%s" is not defined in the XPath context'),
    'XPATH_FUNCTION':
      (2,
       'XPath function "%s" is not defined in the XPath context'),
    'XPATH_FUNC_ARGS':
      (2,
       'XPath function "%s" takes %s arguments but called with %s.'),
    'XPATH_NODE_NOT_FOUND1':
      (4,
       'node "%s::%s" is not found in "%s::%s"'),
    'XPATH_NODE_NOT_FOUND2':
      (4,
       'node "%s::%s" is not found in module "%s"'),
    'XPATH_REF_CONFIG_FALSE':
      (4,
       'node "%s::%s" is config false and is not part of the accessible tree'),
    'XPATH_PATH_TOO_MANY_UP':
      (2,
       'the path has too many ".."'),

#    'XPATH_FUNCTION_RET_VAL':
#      (2,
#       'XPath function "%s" does not return a %s'),
    'AUGMENT_MANDATORY':
      (1,
       'cannot augment with mandatory node "%s"'),

    'LONG_IDENTIFIER':
      (3,
       'identifier "%s" exceeds %s characters'),

    'CONFIG_IGNORED':
      (4,
       'explicit config statement is ignored'),

    'UNUSED_IMPORT':
      (4,
       'imported module "%s" not used'),

    'UNUSED_TYPEDEF':
      (4,
       'locally scoped typedef "%s" not used'),

    'UNUSED_GROUPING':
      (4,
       'locally scoped grouping "%s" not used'),

    'KEY_HAS_DEFAULT':
      (4,
       'default value for a key leaf is ignored'),

    'KEY_HAS_MANDATORY_FALSE':
      (4,
       '"mandatory" statement for a key leaf is ignored'),

    'LONG_LINE':
      (4,
       'line length %s exceeds %s characters'),

    'STRICT_XPATH_FUNCTION':
      (2,
       'XPath function "%s" is not allowed for strict YANG compliance'),
    }

def add_error_code(tag, level, fmt):
    """Add an error code to the framework.

    Can be used by plugins to add special errors."""
    error_codes[tag] = (level, fmt)

def err_level(tag):
    try:
        (level, fmt) = error_codes[tag]
        return level
    except KeyError:
        return 0

def err_to_str(tag, args):
    try:
        (level, fmt) = error_codes[tag]
        return fmt % args
    except KeyError:
        return 'unknown error %s' % tag

def err_add(errors, pos, tag, args):
    error = (copy.copy(pos), tag, args)
    # surely this can be done more elegant??
    for (p, t, a) in errors:
        if (p.line == pos.line and p.ref == pos.ref and
            p.top == pos.top and t == tag and a == args):
            return
    errors.append(error)

def is_warning(level):
    return not is_error(level)

def is_error(level):
    return level < 4

def allow_warning(level):
    return level > 2



"""XPath 1.0 lexer / scanner

Used with the PLY-based parser xpath_parser.py.

See http://www.w3.org/TR/1999/REC-xpath-19991116
"""

import re

class XPathError(Exception):
    def __init__(self, msg, line, pos):
        self.msg = msg
        self.line = line
        self.pos = pos

class XPathTok(object):
    def __init__(self, t, v, line, pos):
        self.type = t
        self.value = v
        self.lineno = line
        self.lexpos = pos

class XPathLexer(object):
    def input(self, s):
        self.toks = []
        self.error = None
        try:
            self.toks = scan(s)
        except SyntaxError as e:
            self.error = e

    def token(self):
        while len(self.toks) > 0:
            tok = self.toks[0]
            self.toks = self.toks[1:]
            if tok.type != '_whitespace':
                return tok

        if self.error is not None:
            raise self.error
        else:
            return None

# not 100% XPath / XML, but good enough for YANG
namestr=r'[a-zA-Z_][a-zA-Z0-9_\-.]*'
ncnamestr = '((' + namestr + '):)?(' + namestr + ')'
prefixteststr = '((' + namestr + r'):)?\*'

re_ncname = re.compile(ncnamestr)

patterns = [
    # special token used when we need to preserve whitespace,
    # not used in normal parsing
    ('_whitespace', re.compile(r'\s+')),
    # Expr tokens
    ('LPAREN', re.compile(r'\(')),
    ('RPAREN', re.compile(r'\)')),
    ('LBRACKET', re.compile(r'\[')),
    ('RBRACKET', re.compile(r'\]')),
    ('DOTDOT', re.compile(r'\.\.')),
    ('DOT', re.compile(r'\.')),
    ('COMMA', re.compile(r'\,')),
    ('AT', re.compile(r'\@')),
    ('DOLLAR', re.compile(r'\$')),
    ('DOUBLECOLON', re.compile(r'::')),
    # operators
    ('DOUBLESLASH', re.compile(r'\/\/')),
    ('SLASH', re.compile(r'\/')),
    ('BAR', re.compile(r'\|')),
    ('PLUS', re.compile(r'\+')),
    ('MINUS', re.compile(r'-')),
    ('EQ', re.compile(r'=')),
    ('NEQ', re.compile(r'!=')),
    ('LTE', re.compile(r'<=')),
    ('GTE', re.compile(r'>=')),
    ('GT', re.compile(r'>')),
    ('LT', re.compile(r'<')),
    ('STAR', re.compile(r'\*')),
    # others
    ('number', re.compile(r'[0-9]+(\.[0-9]+)?')),
    ('prefix_test', re.compile(prefixteststr)),
    ('name', re_ncname),
    ('literal', re.compile(r'(\".*?\")|(\'.*?\')')),
    ]

operators = {
    'div': 'DIV',
    'and': 'AND',
    'or': 'OR',
    'mod': 'MOD',
}

meta_tokens = ['node_type', 'axis', 'function_name', 'wildcard']

node_types = [ 'comment', 'text', 'processing-instruction', 'node' ]
axes = [ 'ancestor-or-self', 'ancestor', 'attribute', 'child',
         'descendant-or-self', 'descendant', 'following-sibling',
         'following', 'namespace', 'parent', 'preceding-sibling',
         'preceding', 'self' ]

def token_defs():
    toks = [p[0] for p in patterns]
    toks.extend(meta_tokens)
    toks.extend([operators[op] for op in operators])
    # remove spc
    toks.remove('_whitespace')
    return toks

re_open_para = re.compile(r'\s*\(')
re_axis = re.compile(r'\s*::')

def scan(s):
    """Return a list of tokens, or throw SyntaxError on failure.
    """
    line = 1
    linepos = 1
    pos = 0
    toks = []
    while pos < len(s):
        matched = False
        for (tokname, r) in patterns:
            m = r.match(s, pos)
            if m is not None:
                # found a matching token
                v = m.group(0)
                prec = _preceding_token(toks)
                if tokname == 'STAR' and prec is not None and _is_special(prec):
                    # XPath 1.0 spec, 3.7 special rule 1a
                    # interpret '*' as a wildcard
                    tok = XPathTok('wildcard', v, line, linepos)
                elif (tokname == 'name' and
                      prec is not None and not _is_special(prec) and
                      v in operators):
                    # XPath 1.0 spec, 3.7 special rule 1b
                    # interpret the name as an operator
                    tok = XPathTok(operators[v], v, line, linepos)
                elif tokname == 'name':
                    # check if next token is '('
                    if re_open_para.match(s, pos + len(v)):
                        # XPath 1.0 spec, 3.7 special rule 2
                        if v in node_types:
                            # XPath 1.0 spec, 3.7 special rule 2a
                            tok = XPathTok('node_type', v, line, linepos)
                        else:
                            # XPath 1.0 spec, 3.7 special rule 2b
                            tok = XPathTok('function_name', v, line, linepos)
                    # check if next token is '::'
                    elif re_axis.match(s, pos + len(v)):
                        # XPath 1.0 spec, 3.7 special rule 3
                        if v in axes:
                            tok = XPathTok('axis', v, line, linepos)
                        else:
                            e = "unknown axis %s" % v
                            raise XPathError(e, line, linepos)
                    else:
                        tok = XPathTok('name', v, line, linepos)
                else:
                    tok = XPathTok(tokname, v, line, linepos)
                if tokname == '_whitespace':
                    n = v.count('\n')
                    if n > 0:
                        line = line + n
                        linepos = len(v) - v.rfind('\n')
                    else:
                        linepos += len(v)
                else:
                    linepos += len(v)
                pos += len(v)
                toks.append(tok)
                matched = True
                break
        if matched == False:
            # no patterns matched
            raise XPathError('syntax error', line, linepos)
    return toks

def _preceding_token(toks):
    if len(toks) > 1 and toks[-1].type == '_whitespace':
        return toks[-2]
    if len(toks) > 0 and toks[-1].type != '_whitespace':
        return toks[-1]
    return None

_special_tok_types = ['AT', 'DOUBLECOLON', 'LPAREN', 'LBRACKET',
                      'SLASH', 'DOUBLESLASH', 'BAR', 'PLUS', 'MINUS',
                      'EQ', 'NEQ', 'LT', 'LTE', 'GT', 'GTE',
                      'AND', 'OR', 'MOD', 'DIV' ]

def _is_special(tok):
    return tok.type in _special_tok_types
